# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from collections.abc import Callable
from typing import Any

import ipywidgets as widgets
import sciline as sl
from IPython import display
from ipywidgets import Layout

from .parameter import Parameter
from .widgets import Spinner, SwitchWidget, create_parameter_widget, default_layout
from .widgets._base import get_fields, set_fields
from .workflow import (
    Key,
    assign_parameter_values,
    get_parameters,
    get_possible_outputs,
    get_typical_outputs,
    workflow_registry,
)


def _wrap_foldable(
    wrapped: widgets.Widget, title: str | None = None
) -> widgets.Accordion:
    return widgets.Accordion(
        [wrapped],
        layout=Layout(width='99%', height='auto'),
        titles=(title,),
    )


class OutputSelectionWidget(widgets.VBox):
    def __init__(self, workflow: sl.Pipeline, **kwargs):
        self.typical_outputs_widget = widgets.SelectMultiple(
            options=get_typical_outputs(workflow),
            layout=Layout(width='90%', height='250px'),
        )
        self.possible_outputs_widget = widgets.SelectMultiple(
            options=get_possible_outputs(workflow),
            layout=Layout(width='90%', height='auto'),
        )
        # Wrapping the selection widget in an individual accordion.
        # It is not possible to have multiple sections open in one ``Accordion`` widget.
        _typical_selection = _wrap_foldable(
            self.typical_outputs_widget, title='Typical Outputs'
        )
        _typical_selection.selected_index = 0  # Open typical outputs by default
        _possible_selection = _wrap_foldable(
            self.possible_outputs_widget, title='Extended Outputs'
        )
        super().__init__([_typical_selection, _possible_selection], **kwargs)

    @property
    def value(self) -> set[Key]:
        return set(
            self.typical_outputs_widget.value + self.possible_outputs_widget.value
        )


class ParameterBox(widgets.VBox):
    def __init__(
        self,
        registry_getter: Callable[[], dict[Key, Parameter]],
        **kwargs,
    ):
        self.parameter_refresh_button = widgets.Button(
            description='Refresh Parameters',
            disabled=False,
            button_style='success',
            tooltip='Generate Parameter Input Widgets',
        )
        self._input_registry = {}
        self._input_widgets = {}
        self._input_box = widgets.VBox()

        def _refresh_input_box(_: widgets.Button):
            new_input_parameters = registry_getter()
            self._input_registry.clear()
            self._input_registry.update(new_input_parameters)
            self._input_widgets.clear()
            self._input_widgets.update(
                {
                    node: create_parameter_widget(parameter)
                    for node, parameter in new_input_parameters.items()
                }
            )
            self._input_box.children = [
                widgets.HBox([widget]) for widget in self._input_widgets.values()
            ]

        self.parameter_refresh_button.on_click(_refresh_input_box)

        super().__init__([self.parameter_refresh_button, self._input_box], **kwargs)

    @property
    def value(self) -> dict[Key, Any]:
        """Return the current parameter values with matching types as a dictionary."""
        return {
            node: widget.value
            for node, widget_box in self._input_widgets.items()
            if (not isinstance((widget := widget_box), SwitchWidget)) or widget.enabled
        }


class ResultBox(widgets.VBox):
    def __init__(
        self,
        workflow_runner: Callable[[], dict[type, Any]],
        result_registry: dict | None = None,
        **kwargs,
    ):
        self.output = widgets.Output()
        self.run_button = widgets.Button(
            description='Run',
            disabled=False,
            button_style='success',
            tooltip='Run',
        )
        output_clear_button = widgets.Button(
            description='Clear Output',
            button_style='warning',
            tooltip='Clear Output',
        )

        def run_workflow(_: widgets.Button) -> None:
            self.output.clear_output()
            with self.output:
                display.display(Spinner())
                compute_result = workflow_runner()
                display.clear_output()
                if result_registry is not None:
                    result_registry.clear()
                    result_registry.update(compute_result)
                for i in compute_result.values():
                    display.display(i)

        def clear_output(_: widgets.Button) -> None:
            self.output.clear_output()

        self.run_button.on_click(run_workflow)
        output_clear_button.on_click(clear_output)
        button_box = widgets.HBox([self.run_button, output_clear_button])
        super().__init__([button_box, self.output], **kwargs)


def connect_refresh_button(
    refresh_button: widgets.Button, output_widget: widgets.Output
) -> None:
    def refresh_output(_: widgets.Button):
        output_widget.clear_output()

    refresh_button.on_click(refresh_output)


def connect_output_selection_and_parameter_run_button(
    *output_selection_widgets: widgets.Widget,
    parameter_refresh_button: widgets.Button,
    run_button: widgets.Button,
) -> None:
    # Disable run button when output selection changes
    def observe_selection_change(_) -> None:
        run_button.disabled = True
        run_button.tooltip = 'To run the workflow, refresh parameters.'

    for output_selection_widget in output_selection_widgets:
        output_selection_widget.observe(observe_selection_change)

    # Enable run button when parameters are generated
    original_run_button_tooltip = run_button.tooltip

    def observe_parameter_refrheshed(_) -> None:
        run_button.disabled = False
        run_button.tooltip = original_run_button_tooltip

    parameter_refresh_button.on_click(observe_parameter_refrheshed)


class WorkflowWidget(widgets.TwoByTwoLayout):
    def __init__(
        self, workflow: sl.Pipeline, result_registry: dict | None = None, **kwargs
    ):
        self.output_selection_box = OutputSelectionWidget(workflow)

        def registry_getter() -> dict[Key, Parameter]:
            """Return the parameter registry for the workflow."""
            return get_parameters(workflow, tuple(self.output_selection_box.value))

        self.parameter_box = ParameterBox(registry_getter)

        def workflow_runner() -> dict[type, Any]:
            """Run the workflow with the current parameter values."""
            return assign_parameter_values(
                workflow.copy(), self.parameter_box.value
            ).compute(self.output_selection_box.value)

        self.result_box = ResultBox(workflow_runner, result_registry)
        connect_refresh_button(
            self.parameter_box.parameter_refresh_button, self.result_box.output
        )
        connect_output_selection_and_parameter_run_button(
            self.output_selection_box.typical_outputs_widget,
            self.output_selection_box.possible_outputs_widget,
            parameter_refresh_button=self.parameter_box.parameter_refresh_button,
            run_button=self.result_box.run_button,
        )
        for box in (self.output_selection_box, self.parameter_box, self.result_box):
            box.layout.border = '1px solid black'

        super().__init__(
            top_left=self.output_selection_box,
            top_right=self.parameter_box,
            bottom_left=self.result_box,
            grid_gap="10px",
            layout=default_layout,
            **kwargs,
        )


def workflow_widget(result_registry: dict | None = None) -> widgets.Widget:
    """Create a widget for a workflow selected from a dropdown."""
    workflow_select = widgets.Dropdown(
        options=[(workflow.__name__, workflow) for workflow in workflow_registry],
        description='Workflow:',
        value=None,
        layout=default_layout,
        tooltip='Select a workflow.',
    )

    def refresh_workflow_box(change) -> None:
        workflow_box.children = [WorkflowWidget(change.new(), result_registry)]

    workflow_select.observe(refresh_workflow_box, names='value')

    workflow_selection_box = widgets.HBox([workflow_select], layout=default_layout)
    workflow_box = widgets.Box(layout=default_layout)
    return widgets.VBox([workflow_selection_box, workflow_box])


def _get_parameter_box(widget: WorkflowWidget | ParameterBox) -> ParameterBox:
    if isinstance(widget, WorkflowWidget):
        return widget.parameter_box
    elif isinstance(widget, ParameterBox):
        return widget
    else:
        raise TypeError(
            f"Expected target_widget to be a WorkflowWidget or ParameterBox, "
            f"got {type(widget)}."
        )


def set_parameter_widget_values(
    widget: WorkflowWidget | ParameterBox, new_parameter_values: dict[type, Any]
) -> None:
    """Set the values of the input widgets in the target widget.

    Nodes that don't exist in the input widgets will be ignored.

    Example
    -------
    .. code-block::

        set_parameter_widget_values(widget, {
            'WavelengthBins': {'start': 1.0, 'stop': 14.0, 'nbins': 500}
        })

    Parameters
    ----------
    widget:
        The widget containing the input widgets.
    new_parameter_values:
        A dictionary of values/state to set each fields/state or value of input widgets.

    Raises
    ------
    TypeError:
        If the widget is not a WorkflowWidget or a ParameterBox.

    """
    parameter_box = _get_parameter_box(widget)
    # Walk through the existing input widgets and set the values
    # ``node`s that don't exist in the input widgets will be ignored.
    for node, widget in parameter_box._input_widgets.items():
        if node in new_parameter_values:
            # We shouldn't use `get` here because ``None`` is a valid value.
            set_fields(widget, new_parameter_values[node])


def get_parameter_widget_values(
    widget: WorkflowWidget | ParameterBox,
) -> dict[type, Any]:
    """Return the current values of the input widgets in the target widget.

    The result of this function can be used to set the values of the input widgets
    using the :py:func:`~set_parameter_widget_values` function.

    Parameters
    ----------
    widget:
        The widget containing the input widgets.

    Returns
    -------
    :
        A dictionary of the current values/state of each input widget.

    Raises
    ------
    TypeError:
        If the widget is not a WorkflowWidget or a ParameterBox.

    """
    return {
        node: get_fields(widget)
        for node, widget in _get_parameter_box(widget)._input_widgets.items()
    }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

import doctest
import os
import sys
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as get_version

from sphinx.util import logging

sys.path.insert(0, os.path.abspath('.'))

logger = logging.getLogger(__name__)

# General information about the project.
project = 'ESSreduce'
copyright = '2024 Scipp contributors'
author = 'Scipp contributors'

html_show_sourcelink = True

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_copybutton',
    'sphinx_design',
    'nbsphinx',
    'myst_parser',
]

try:
    import sciline.sphinxext.domain_types  # noqa: F401

    extensions.append('sciline.sphinxext.domain_types')
    # See https://github.com/tox-dev/sphinx-autodoc-typehints/issues/457
    suppress_warnings = ["config.cache"]
except ModuleNotFoundError:
    pass


myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "fieldlist",
    "html_admonition",
    "html_image",
    "replacements",
    "smartquotes",
    "strikethrough",
    "substitution",
    "tasklist",
]

myst_heading_anchors = 3

autodoc_type_aliases = {
    'array_like': 'array_like',
}

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'numpy': ('https://numpy.org/doc/stable/', None),
    'scipp': ('https://scipp.github.io/', None),
}

# autodocs includes everything, even irrelevant API internals. autosummary
# looks more suitable in the long run when the API grows.
# For a nice example see how xarray handles its API documentation.
autosummary_generate = True

napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = True
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # objects without namespace: numpy
    "ndarray": "~numpy.ndarray",
}
typehints_defaults = 'comma'
typehints_use_rtype = False


sciline_domain_types_prefix = 'ess.reduce'
sciline_domain_types_aliases = {
    'scipp._scipp.core.DataArray': 'scipp.DataArray',
    'scipp._scipp.core.Dataset': 'scipp.Dataset',
    'scipp._scipp.core.DType': 'scipp.DType',
    'scipp._scipp.core.Unit': 'scipp.Unit',
    'scipp._scipp.core.Variable': 'scipp.Variable',
    'scipp.core.data_group.DataGroup': 'scipp.DataGroup',
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
html_sourcelink_suffix = ''  # Avoid .ipynb.txt extensions in sources

# The master toctree document.
master_doc = 'index'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#

try:
    release = get_version("essreduce")
    version = ".".join(release.split('.')[:3])  # CalVer
except PackageNotFoundError:
    logger.info(
        "Warning: determining version from package metadata failed, falling back to "
        "a dummy version number."
    )
    release = version = "0.0.0-dev"

warning_is_error = True

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', '**.ipynb_checkpoints']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "primary_sidebar_end": ["edit-this-page", "sourcelink"],
    "secondary_sidebar_items": [],
    "navbar_persistent": ["search-button"],
    "show_nav_level": 1,
    # Adjust this to ensure external links are moved to "Move" menu
    "header_links_before_dropdown": 4,
    "pygment_light_style": "github-light-high-contrast",
    "pygment_dark_style": "github-dark-high-contrast",
    "logo": {
        "image_light": "_static/logo.svg",
        "image_dark": "_static/logo-dark.svg",
    },
    "external_links": [
        {"name": "ESSdiffraction", "url": "https://scipp.github.io/essdiffraction"},
        {"name": "ESSimaging", "url": "https://scipp.github.io/essimaging"},
        {"name": "ESSnmx", "url": "https://scipp.github.io/essnmx"},
        {"name": "ESSpolarization", "url": "https://scipp.github.io/esspolarization"},
        {"name": "ESSreflectometry", "url": "https://scipp.github.io/essreflectometry"},
        {"name": "ESSsans", "url": "https://scipp.github.io/esssans"},
        {"name": "ESSspectroscopy", "url": "https://scipp.github.io/essspectroscopy"},
    ],
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/scipp/essreduce",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/essreduce/",
            "icon": "fa-brands fa-python",
            "type": "fontawesome",
        },
        {
            "name": "Conda",
            "url": "https://anaconda.org/scipp/essreduce",
            "icon": "fa-custom fa-anaconda",
            "type": "fontawesome",
        },
    ],
    "footer_start": ["copyright", "sphinx-version"],
    "footer_end": ["doc_version", "theme-version"],
}
html_context = {
    "doc_path": "docs",
}
html_sidebars = {
    "**": ["sidebar-nav-bs", "page-toc"],
}

html_title = "ESSreduce"
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.svg"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = []
html_js_files = ["anaconda-icon.js"]

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'essreducedoc'

# -- Options for Matplotlib in notebooks ----------------------------------

nbsphinx_execute_arguments = [
    "--Session.metadata=scipp_sphinx_build=True",
]

# -- Options for doctest --------------------------------------------------

# sc.plot returns a Figure object and doctest compares that against the
# output written in the docstring. But we only want to show an image of the
# figure, not its `repr`.
# In addition, there is no need to make plots in doctest as the documentation
# build already tests if those plots can be made.
# So we simply disable plots in doctests.
doctest_global_setup = '''
import numpy as np

try:
    import scipp as sc

    def do_not_plot(*args, **kwargs):
        pass

    sc.plot = do_not_plot
    sc.Variable.plot = do_not_plot
    sc.DataArray.plot = do_not_plot
    sc.DataGroup.plot = do_not_plot
    sc.Dataset.plot = do_not_plot
except ImportError:
    # Scipp is not needed by docs if it is not installed.
    pass
'''

# Using normalize whitespace because many __str__ functions in scipp produce
# extraneous empty lines and it would look strange to include them in the docs.
doctest_default_flags = (
    doctest.ELLIPSIS
    | doctest.IGNORE_EXCEPTION_DETAIL
    | doctest.DONT_ACCEPT_TRUE_FOR_1
    | doctest.NORMALIZE_WHITESPACE
)

# -- Options for linkcheck ------------------------------------------------

linkcheck_ignore = [
    # Specific lines in Github blobs cannot be found by linkcheck.
    r'https?://github\.com/.*?/blob/[a-f0-9]+/.+?#',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

"""
Utilities for computing real neutron time-of-flight from chopper settings and
neutron time-of-arrival at the detectors.
"""

from .simulation import simulate_beamline
from .toa_to_tof import default_parameters, resample_tof_data, providers, TofWorkflow
from .to_events import to_events
from .types import (
    DistanceResolution,
    FrameFoldedTimeOfArrival,
    FramePeriod,
    LookupTableRelativeErrorThreshold,
    Ltotal,
    LtotalRange,
    MaskedTimeOfFlightLookupTable,
    PivotTimeAtDetector,
    PulsePeriod,
    PulseStride,
    PulseStrideOffset,
    RawData,
    ResampledTofData,
    SimulationResults,
    TimeOfArrivalMinusPivotTimeModuloPeriod,
    TimeOfFlightLookupTable,
    TofData,
    UnwrappedTimeOfArrival,
    UnwrappedTimeOfArrivalMinusPivotTime,
)


__all__ = [
    "DistanceResolution",
    "FrameFoldedTimeOfArrival",
    "FramePeriod",
    "LookupTableRelativeErrorThreshold",
    "Ltotal",
    "LtotalRange",
    "MaskedTimeOfFlightLookupTable",
    "PivotTimeAtDetector",
    "PulsePeriod",
    "PulseStride",
    "PulseStrideOffset",
    "RawData",
    "ResampledTofData",
    "SimulationResults",
    "TimeOfArrivalMinusPivotTimeModuloPeriod",
    "TimeOfFlightLookupTable",
    "TofData",
    "TofWorkflow",
    "UnwrappedTimeOfArrival",
    "UnwrappedTimeOfArrivalMinusPivotTime",
    "default_parameters",
    "providers",
    "resample_tof_data",
    "simulate_beamline",
    "to_events",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc

from ess.reduce.live import raw


def test_clear_counts_resets_counts_to_zero() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.Detector(detector_number)
    assert det.data.sum().value == 0
    det.add_counts([1, 2, 3, 2])
    assert det.data.sum().value == 4
    det.clear_counts()
    assert det.data.sum().value == 0


def test_Detector_bincount_drops_out_of_range_ids() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.Detector(detector_number)
    counts = det.bincount([1, 2, 0, -1, 3, 4])
    assert sc.identical(
        counts.data,
        sc.array(dims=['pixel'], values=[1, 1, 1], unit='counts', dtype='int32'),
    )


def test_RollingDetectorView_full_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=2)
    rolling = det.get()
    assert rolling.sizes == {'pixel': 3}
    assert rolling.sum().value == 0
    det.add_counts([1, 2, 3, 2])
    assert det.get().sum().value == 4
    det.add_counts([1, 3, 2])
    assert det.get().sum().value == 7
    det.add_counts([1, 2])
    assert det.get().sum().value == 5
    det.add_counts([])
    assert det.get().sum().value == 2


def test_RollingDetectorView_partial_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=3)
    det.add_counts([1, 2, 3, 2])
    assert det.get(0).sum().value == 0
    assert det.get(1).sum().value == 4
    assert det.get(2).sum().value == 4
    assert det.get(3).sum().value == 4
    det.add_counts([1, 3, 2])
    assert det.get(0).sum().value == 0
    assert det.get(1).sum().value == 3
    assert det.get(2).sum().value == 7
    assert det.get(3).sum().value == 7
    det.add_counts([1, 2])
    assert det.get(0).sum().value == 0
    assert det.get(1).sum().value == 2
    assert det.get(2).sum().value == 5
    assert det.get(3).sum().value == 9
    det.add_counts([])
    assert det.get(0).sum().value == 0
    assert det.get(1).sum().value == 0
    assert det.get(2).sum().value == 2
    assert det.get(3).sum().value == 5
    det.add_counts([1, 2])
    assert det.get(0).sum().value == 0
    assert det.get(1).sum().value == 2
    assert det.get(2).sum().value == 2
    assert det.get(3).sum().value == 4


def test_RollingDetectorView_raises_if_subwindow_exceeds_window() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    det = raw.RollingDetectorView(detector_number=detector_number, window=3)
    with pytest.raises(ValueError, match="Window size"):
        det.get(4)
    with pytest.raises(ValueError, match="Window size"):
        det.get(-1)


def test_RollingDetectorView_projects_counts() -> None:
    detector_number = sc.array(dims=['pixel'], values=[1, 2, 3], unit=None)
    # Dummy projection that just drops the first pixel
    det = raw.RollingDetectorView(
        detector_number=detector_number,
        window=3,
        projection=lambda da: da['pixel', 1:].rename(pixel='abc'),
    )
    expected = sc.DataArray(
        sc.array(dims=['pixel'], values=[0, 0], unit='counts', dtype='int32'),
        coords={'detector_number': detector_number[1:]},
    ).rename(pixel='abc')

    det.add_counts([1, 2, 3, 2])
    expected.values = [2, 1]
    assert sc.identical(det.get(), expected)

    det.add_counts([1, 3, 3, 1])
    expected.values = [2, 3]
    assert sc.identical(det.get(), expected)


def test_project_xy_with_given_zplane_scales() -> None:
    result = raw.project_xy(
        sc.vectors(
            dims=['point'],
            values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [6.0, 10.0, 12.0]],
            unit='m',
        ),
        zplane=sc.scalar(6.0, unit='m'),
    )
    assert sc.identical(
        result,
        sc.DataGroup(
            x=sc.array(dims=['point'], values=[2.0, 4.0, 3.0], unit='m'),
            y=sc.array(dims=['point'], values=[4.0, 5.0, 5.0], unit='m'),
            z=sc.scalar(6.0, unit='m'),
        ),
    )


def test_project_xy_defaults_to_scale_to_zmin() -> None:
    result = raw.project_xy(
        sc.vectors(
            dims=['point'],
            values=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [6.0, 10.0, 12.0]],
            unit='m',
        )
    )
    assert sc.identical(
        result,
        sc.DataGroup(
            # Note same relative values as with zplane=6, just scaled.
            x=sc.array(dims=['point'], values=[1.0, 2.0, 1.5], unit='m'),
            y=sc.array(dims=['point'], values=[2.0, 2.5, 2.5], unit='m'),
            z=sc.scalar(3.0, unit='m'),
        ),
    )


def test_project_onto_cylinder_z() -> None:
    radius = sc.scalar(2.0, unit='m')
    # Input radii are 4 and 1 => scale by 1/2 and 2.
    result = raw.project_onto_cylinder_z(
        sc.vectors(dims=['point'], values=[[0.0, 4.0, 3.0], [1.0, 0.0, 6.0]], unit='m'),
        radius=radius,
    )
    assert sc.identical(result['r'], radius)
    assert sc.identical(
        result['z'], sc.array(dims=['point'], values=[1.5, 12.0], unit='m')
    )
    assert sc.identical(
        result['phi'], sc.array(dims=['point'], values=[90.0, 0.0], unit='deg')
    )
    assert sc.identical(
        result['arc_length'],
        sc.array(dims=['point'], values=[radius.value * np.pi * 0.5, 0.0], unit='m'),
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from typing import Any

from ipywidgets import Checkbox, HBox, Label, Stack, Widget

from ._base import get_fields, set_fields
from ._config import default_style


class SwitchWidget(HBox):
    """Wrapper widget to handle switchable widgets.

    When you retrieve the value of this widget,
    it will return the value of the wrapped widget.
    It is expected not to be set in the workflow if ``enabled`` is False.
    """

    def __init__(self, wrapped: Widget, name: str = '') -> None:
        super().__init__()
        self._enable_box = Checkbox(description='', style=default_style)
        # The layout is not applied if they are set in the constructor
        self._enable_box.layout.description_width = '0px'
        self._enable_box.layout.width = 'auto'

        self.wrapped = wrapped
        wrapped_stack = Stack([Label(name, style=default_style), self.wrapped])
        # We wanted to implement this by greying out the widget when disabled
        # but ``disabled`` is not a common property of all widgets
        wrapped_stack.selected_index = 0

        def handle_checkbox(change) -> None:
            wrapped_stack.selected_index = 1 if change.new else 0

        self._enable_box.observe(handle_checkbox, names='value')
        self.children = [self._enable_box, wrapped_stack]

    @property
    def enabled(self) -> bool:
        return self._enable_box.value

    @enabled.setter
    def enabled(self, value: bool) -> None:
        self._enable_box.value = value

    @property
    def value(self) -> Any:
        return self.wrapped.value

    @value.setter
    def value(self, value: Any) -> None:
        self.wrapped.value = value

    def set_fields(self, new_values: dict[str, Any]) -> None:
        # Retrieve and set the enabled flag first
        new_values = dict(new_values)
        enabled_flag = new_values.pop('enabled', self.enabled)
        if not isinstance(enabled_flag, bool):
            raise ValueError(f"`enabled` must be a boolean, got {enabled_flag}")
        self.enabled = enabled_flag
        # Set the rest of the fields
        set_fields(self.wrapped, new_values)

    def get_fields(self) -> dict[str, Any]:
        wrapped_fields = get_fields(self.wrapped)
        return {'enabled': self.enabled, **(wrapped_fields or {})}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from typing import NewType

import sciline
import scipp as sc

from ess.reduce import streaming


def test_eternal_accumulator_sums_everything() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.sum(var))


def test_eternal_accumulator_sums_everything_with_preprocess() -> None:
    accum = streaming.EternalAccumulator(preprocess=lambda x: x**0.5)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    assert sc.identical(accum.value, sc.sum(var**0.5))


def test_eternal_accumulator_works_if_output_value_is_modified() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    value = accum.value
    value += 1.0
    assert sc.identical(accum.value, sc.sum(var))


def test_eternal_accumulator_does_not_modify_pushed_values() -> None:
    accum = streaming.EternalAccumulator()
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    original = var.copy()
    for i in range(10):
        accum.push(var[i])
    assert sc.identical(var, original)


def test_rolling_accumulator_sums_over_window() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert sc.identical(accum.value, var[0])
    accum.push(var[1].copy())
    assert sc.identical(accum.value, var[0:2].sum())
    accum.push(var[2].copy())
    assert sc.identical(accum.value, var[0:3].sum())
    accum.push(var[3].copy())
    assert sc.identical(accum.value, var[1:4].sum())
    accum.push(var[4].copy())
    assert sc.identical(accum.value, var[2:5].sum())


def test_rolling_accumulator_sums_over_window_with_preprocess() -> None:
    accum = streaming.RollingAccumulator(window=3, preprocess=lambda x: x**0.5)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    accum.push(var[0].copy())
    assert sc.identical(accum.value, var[0] ** 0.5)
    accum.push(var[1].copy())
    assert sc.identical(accum.value, (var[0:2] ** 0.5).sum())
    accum.push(var[2].copy())
    assert sc.identical(accum.value, (var[0:3] ** 0.5).sum())
    accum.push(var[3].copy())
    assert sc.identical(accum.value, (var[1:4] ** 0.5).sum())
    accum.push(var[4].copy())
    assert sc.identical(accum.value, (var[2:5] ** 0.5).sum())


def test_rolling_accumulator_works_if_output_value_is_modified() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    for i in range(10):
        accum.push(var[i].copy())
    value = accum.value
    value += 1.0
    assert sc.identical(accum.value, var[7:10].sum())


def test_rolling_accumulator_does_not_modify_pushed_values() -> None:
    accum = streaming.RollingAccumulator(window=3)
    var = sc.linspace(dim='x', start=0, stop=1, num=10)
    original = var.copy()
    for i in range(10):
        accum.push(var[i])
    assert sc.identical(var, original)


DynamicA = NewType('DynamicA', float)
DynamicB = NewType('DynamicB', float)
StaticA = NewType('StaticA', float)
AccumA = NewType('AccumA', float)
AccumB = NewType('AccumB', float)
Target = NewType('Target', float)


def make_static_a() -> StaticA:
    make_static_a.call_count += 1
    return StaticA(2.0)


make_static_a.call_count = 0


def make_accum_a(value: DynamicA, static: StaticA) -> AccumA:
    return AccumA(value * static)


def make_accum_b(value: DynamicB) -> AccumB:
    return AccumB(value)


def make_target(accum_a: AccumA, accum_b: AccumB) -> Target:
    return Target(accum_a / accum_b)


def test_StreamProcessor_overall_behavior() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )
    orig_workflow = base_workflow.copy()

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 9.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 15.0))
    assert make_static_a.call_count == 1

    # Consistency check: Run the original workflow with the same inputs, all at once
    orig_workflow[DynamicA] = sc.scalar(1 + 2 + 3)
    orig_workflow[DynamicB] = sc.scalar(4 + 5 + 6)
    expected = orig_workflow.compute(Target)
    assert sc.identical(expected, result[Target])


def test_StreamProcessor_uses_custom_accumulator() -> None:
    class Always42Accumulator(streaming.Accumulator[sc.Variable]):
        def _do_push(self, value: sc.Variable) -> None:
            pass

        @property
        def value(self) -> sc.Variable:
            return sc.scalar(42)

    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators={
            AccumA: streaming.EternalAccumulator(),
            AccumB: Always42Accumulator(),
        },
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 42.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 42.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 6.0 / 42.0))


def test_StreamProcessor_does_not_compute_unused_static_nodes() -> None:
    def a_independent_target(accum_b: AccumB) -> Target:
        return Target(1.5 * accum_b)

    def derived_static_a(x: float) -> StaticA:
        derived_static_a.call_count += 1
        return StaticA(2.0 * x)

    derived_static_a.call_count = 0

    base_workflow = sciline.Pipeline(
        (derived_static_a, make_accum_a, make_accum_b, a_independent_target)
    )

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(AccumA, AccumB),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(1.5 * 4.0))
    assert derived_static_a.call_count == 0


def test_StreamProcess_with_zero_accumulators_for_buffered_workflow_calls() -> None:
    base_workflow = sciline.Pipeline(
        (make_static_a, make_accum_a, make_accum_b, make_target)
    )
    make_static_a.call_count = 0

    streaming_wf = streaming.StreamProcessor(
        base_workflow=base_workflow,
        dynamic_keys=(DynamicA, DynamicB),
        target_keys=(Target,),
        accumulators=(),
    )
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(1), DynamicB: sc.scalar(4)})
    assert sc.identical(result[Target], sc.scalar(2 * 1.0 / 4.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(2), DynamicB: sc.scalar(5)})
    assert sc.identical(result[Target], sc.scalar(2 * 2.0 / 5.0))
    result = streaming_wf.add_chunk({DynamicA: sc.scalar(3), DynamicB: sc.scalar(6)})
    assert sc.identical(result[Target], sc.scalar(2 * 3.0 / 6.0))
    assert make_static_a.call_count == 1



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""Live data utilities"""

from .workflow import LiveWorkflow

__all__ = [
    'LiveWorkflow',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)

from functools import reduce

import numpy as np
import scipp as sc


def to_events(
    da: sc.DataArray, event_dim: str, events_per_bin: int = 500
) -> sc.DataArray:
    """
    Convert a histogrammed data array to an event list.
    The generated events have a uniform distribution within each bin.
    Each dimension with a bin-edge coordinate is converted to an event coordinate.
    The contract is that if we re-histogram the event list with the same bin edges,
    we should get the original counts back.
    Masks on non-bin-edge dimensions are preserved.
    If there are masks on bin-edge dimensions, the masked values are zeroed out in the
    original data before the conversion to events.

    Parameters
    ----------
    da:
        DataArray to convert to events.
    event_dim:
        Name of the new event dimension.
    events_per_bin:
        Number of events to generate per bin.
    """
    if da.bins is not None:
        raise ValueError("Cannot convert a binned DataArray to events.")
    rng = np.random.default_rng()
    event_coords = {}
    edge_dims = []
    midp_dims = []
    # Separate bin-edge and midpoints coords
    for dim in da.dims:
        if da.coords.is_edges(dim):
            edge_dims.append(dim)
        else:
            midp_dims.append(dim)

    edge_sizes = {dim: da.sizes[dim] for dim in edge_dims}
    for dim in edge_dims:
        coord = da.coords[dim]
        left = sc.broadcast(coord[dim, :-1], sizes=edge_sizes).values
        right = sc.broadcast(coord[dim, 1:], sizes=edge_sizes).values

        # The numpy.random.uniform function below does not support NaNs, so we need to
        # replace them with zeros, and then replace them back after the random numbers
        # have been generated.
        nans = np.isnan(left) | np.isnan(right)
        left = np.where(nans, 0.0, left)
        right = np.where(nans, 0.0, right)
        # Ensure left <= right
        left, right = np.minimum(left, right), np.maximum(left, right)

        # In each bin, we generate a number of events with a uniform distribution.
        events = rng.uniform(
            left, right, size=(events_per_bin, *list(edge_sizes.values()))
        )
        events[..., nans] = np.nan
        event_coords[dim] = sc.array(
            dims=[event_dim, *edge_dims], values=events, unit=coord.unit
        )

    # Find and apply masks that are on a bin-edge dimension
    event_masks = {}
    other_masks = {}
    edge_dims_set = set(edge_dims)
    for key, mask in da.masks.items():
        if set(mask.dims) & edge_dims_set:
            event_masks[key] = mask
        else:
            other_masks[key] = mask

    data = da.data
    if event_masks:
        inv_mask = (~reduce(lambda a, b: a | b, event_masks.values())).to(dtype=int)
        inv_mask.unit = ""
        data = data * inv_mask

    # Create the data counts, which are the original counts divided by the number of
    # events per bin
    sizes = {event_dim: events_per_bin} | da.sizes
    val = sc.broadcast(sc.values(data) / float(events_per_bin), sizes=sizes)
    kwargs = {"dims": sizes.keys(), "values": val.values, "unit": data.unit}
    if data.variances is not None:
        # Note here that all the events are correlated.
        # If we later histogram the events with different edges than the original
        # histogram, then neighboring bins will be correlated, and the error obtained
        # will be too small. It is however not clear what can be done to improve this.
        kwargs["variances"] = sc.broadcast(
            sc.variances(data) / float(events_per_bin), sizes=sizes
        ).values
    new_data = sc.array(**kwargs)

    new = sc.DataArray(data=new_data, coords=event_coords)
    new = new.transpose((*midp_dims, *edge_dims, event_dim)).flatten(
        dims=[*edge_dims, event_dim], to=event_dim
    )
    return new.assign_coords(
        {dim: da.coords[dim].copy() for dim in midp_dims}
    ).assign_masks({key: mask.copy() for key, mask in other_masks.items()})



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
from ipywidgets import HBox, Text, ValueWidget

from ._config import default_layout


class StringWidget(HBox, ValueWidget):
    def __init__(self, description: str, value: str | None = None, **kwargs):
        super().__init__(layout=default_layout)
        self.text_widget = Text(description=description, value=value, **kwargs)
        self.children = [self.text_widget]

    @property
    def value(self) -> str | None:
        v = self.text_widget.value.strip()
        if not v:
            return None
        return v

    @value.setter
    def value(self, value: str | None):
        if value is None:
            self.text_widget.value = ''
        else:
            self.text_widget.value = value


class MultiStringWidget(StringWidget):
    def __init__(
        self, description: str, value: tuple[str, ...] | None = None, **kwargs
    ):
        # Special case handling to allow initialising with a single string
        if not isinstance(value, str) and value is not None:
            value = ', '.join(value)

        super().__init__(description, value, **kwargs)

    @property
    def value(self) -> tuple[str, ...]:
        v = super().value
        if v is None:
            return ()
        return tuple(s.strip() for s in v.split(','))

    @value.setter
    def value(self, value: tuple[str, ...]):
        self.text_widget.value = ', '.join(value)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
Time-of-flight workflow for unwrapping the time of arrival of the neutron at the
detector.
This workflow is used to convert raw detector data with event_time_zero and
event_time_offset coordinates to data with a time-of-flight coordinate.
"""

from collections.abc import Callable

import numpy as np
import scipp as sc
from scipp._scipp.core import _bins_no_validate
from scippneutron._utils import elem_unit

from .to_events import to_events
from .types import (
    DistanceResolution,
    FastestNeutron,
    FrameFoldedTimeOfArrival,
    FramePeriod,
    LookupTableRelativeErrorThreshold,
    Ltotal,
    LtotalRange,
    MaskedTimeOfFlightLookupTable,
    PivotTimeAtDetector,
    PulsePeriod,
    PulseStride,
    PulseStrideOffset,
    RawData,
    ResampledTofData,
    SimulationResults,
    TimeOfArrivalMinusPivotTimeModuloPeriod,
    TimeOfArrivalResolution,
    TimeOfFlightLookupTable,
    TofData,
    UnwrappedTimeOfArrival,
    UnwrappedTimeOfArrivalMinusPivotTime,
)


def frame_period(pulse_period: PulsePeriod, pulse_stride: PulseStride) -> FramePeriod:
    """
    Return the period of a frame, which is defined by the pulse period times the pulse
    stride.

    Parameters
    ----------
    pulse_period:
        Period of the source pulses, i.e., time between consecutive pulse starts.
    pulse_stride:
        Stride of used pulses. Usually 1, but may be a small integer when
        pulse-skipping.
    """
    return FramePeriod(pulse_period * pulse_stride)


def extract_ltotal(da: RawData) -> Ltotal:
    """
    Extract the total length of the flight path from the source to the detector from the
    detector data.

    Parameters
    ----------
    da:
        Raw detector data loaded from a NeXus file, e.g., NXdetector containing
        NXevent_data.
    """
    return Ltotal(da.coords["Ltotal"])


def compute_tof_lookup_table(
    simulation: SimulationResults,
    ltotal_range: LtotalRange,
    distance_resolution: DistanceResolution,
    toa_resolution: TimeOfArrivalResolution,
) -> TimeOfFlightLookupTable:
    """
    Compute a lookup table for time-of-flight as a function of distance and
    time-of-arrival.

    Parameters
    ----------
    simulation:
        Results of a time-of-flight simulation used to create a lookup table.
        The results should be a flat table with columns for time-of-arrival, speed,
        wavelength, and weight.
    ltotal_range:
        Range of total flight path lengths from the source to the detector.
    distance_resolution:
        Resolution of the distance axis in the lookup table.
    toa_resolution:
        Resolution of the time-of-arrival axis in the lookup table.
    """
    distance_unit = "m"
    res = distance_resolution.to(unit=distance_unit)
    simulation_distance = simulation.distance.to(unit=distance_unit)

    min_dist, max_dist = (
        x.to(unit=distance_unit) - simulation_distance for x in ltotal_range
    )
    # We need to bin the data below, to compute the weighted mean of the wavelength.
    # This results in data with bin edges.
    # However, the 2d interpolator expects bin centers.
    # We want to give the 2d interpolator a table that covers the requested range,
    # hence we need to extend the range by at least half a resolution in each direction.
    # Then, we make the choice that the resolution in distance is the quantity that
    # should be preserved. Because the difference between min and max distance is
    # not necessarily an integer multiple of the resolution, we need to add a pad to
    # ensure that the last bin is not cut off. We want the upper edge to be higher than
    # the maximum distance, hence we pad with an additional 1.5 x resolution.
    pad = 2.0 * res
    dist_edges = sc.array(
        dims=["distance"],
        values=np.arange((min_dist - pad).value, (max_dist + pad).value, res.value),
        unit=distance_unit,
    )
    distances = sc.midpoints(dist_edges)

    time_unit = simulation.time_of_arrival.unit
    toas = simulation.time_of_arrival + (distances / simulation.speed).to(
        unit=time_unit, copy=False
    )

    # Compute time-of-flight for all neutrons
    wavs = sc.broadcast(simulation.wavelength.to(unit="m"), sizes=toas.sizes).flatten(
        to="event"
    )
    dist = sc.broadcast(distances + simulation_distance, sizes=toas.sizes).flatten(
        to="event"
    )
    tofs = dist * sc.constants.m_n
    tofs *= wavs
    tofs /= sc.constants.h

    data = sc.DataArray(
        data=sc.broadcast(simulation.weight, sizes=toas.sizes).flatten(to="event"),
        coords={
            "toa": toas.flatten(to="event"),
            "tof": tofs.to(unit=time_unit, copy=False),
            "distance": dist,
        },
    )

    binned = data.bin(distance=dist_edges + simulation_distance, toa=toa_resolution)
    # Weighted mean of tof inside each bin
    mean_tof = (
        binned.bins.data * binned.bins.coords["tof"]
    ).bins.sum() / binned.bins.sum()
    # Compute the variance of the tofs to track regions with large uncertainty
    variance = (
        binned.bins.data * (binned.bins.coords["tof"] - mean_tof) ** 2
    ).bins.sum() / binned.bins.sum()

    mean_tof.variances = variance.values

    # Convert coordinates to midpoints
    mean_tof.coords["toa"] = sc.midpoints(mean_tof.coords["toa"])
    mean_tof.coords["distance"] = sc.midpoints(mean_tof.coords["distance"])

    return TimeOfFlightLookupTable(mean_tof)


def masked_tof_lookup_table(
    tof_lookup: TimeOfFlightLookupTable,
    error_threshold: LookupTableRelativeErrorThreshold,
) -> MaskedTimeOfFlightLookupTable:
    """
    Mask regions of the lookup table where the variance of the projected time-of-flight
    is larger than a given threshold.

    Parameters
    ----------
    tof_lookup:
        Lookup table giving time-of-flight as a function of distance and
        time-of-arrival.
    variance_threshold:
        Threshold for the variance of the projected time-of-flight above which regions
        are masked.
    """
    relative_error = sc.stddevs(tof_lookup.data) / sc.values(tof_lookup.data)
    mask = relative_error > sc.scalar(error_threshold)
    out = tof_lookup.copy()
    # Use numpy for indexing as table is 2D
    out.values[mask.values] = np.nan
    return MaskedTimeOfFlightLookupTable(out)


def find_fastest_neutron(simulation: SimulationResults) -> FastestNeutron:
    """
    Find the fastest neutron in the simulation results.
    """
    ind = np.argmax(simulation.speed.values)
    return FastestNeutron(
        time_of_arrival=simulation.time_of_arrival[ind],
        speed=simulation.speed[ind],
        distance=simulation.distance,
    )


def pivot_time_at_detector(
    fastest_neutron: FastestNeutron, ltotal: Ltotal
) -> PivotTimeAtDetector:
    """
    Compute the pivot time at the detector, i.e., the time of the start of the frame at
    the detector.
    The assumption here is that the fastest neutron in the simulation results is the one
    that arrives at the detector first.
    One could have an edge case where a slightly slower neutron which is born earlier
    could arrive at the detector first, but this edge case is most probably uncommon,
    and the difference in arrival times is likely to be small.

    Parameters
    ----------
    fastest_neutron:
        Properties of the fastest neutron in the simulation results.
    ltotal:
        Total length of the flight path from the source to the detector.
    """
    dist = ltotal - fastest_neutron.distance.to(unit=ltotal.unit)
    toa = fastest_neutron.time_of_arrival + (dist / fastest_neutron.speed).to(
        unit=fastest_neutron.time_of_arrival.unit, copy=False
    )
    return PivotTimeAtDetector(toa)


def unwrapped_time_of_arrival(
    da: RawData, offset: PulseStrideOffset, pulse_period: PulsePeriod
) -> UnwrappedTimeOfArrival:
    """
    Compute the unwrapped time of arrival of the neutron at the detector.
    For event data, this is essentially ``event_time_offset + event_time_zero``.

    Parameters
    ----------
    da:
        Raw detector data loaded from a NeXus file, e.g., NXdetector containing
        NXevent_data.
    offset:
        Integer offset of the first pulse in the stride (typically zero unless we are
        using pulse-skipping and the events do not begin with the first pulse in the
        stride).
    pulse_period:
        Period of the source pulses, i.e., time between consecutive pulse starts.
    """
    if da.bins is None:
        # 'time_of_flight' is the canonical name in NXmonitor, but in some files, it
        # may be called 'tof'.
        key = next(iter(set(da.coords.keys()) & {"time_of_flight", "tof"}))
        toa = da.coords[key]
    else:
        # To unwrap the time of arrival, we want to add the event_time_zero to the
        # event_time_offset. However, we do not really care about the exact datetimes,
        # we just want to know the offsets with respect to the start of the run.
        # Hence we use the smallest event_time_zero as the time origin.
        time_zero = da.coords["event_time_zero"] - da.coords["event_time_zero"].min()
        coord = da.bins.coords["event_time_offset"]
        unit = elem_unit(coord)
        toa = (
            coord
            + time_zero.to(dtype=float, unit=unit, copy=False)
            - (offset * pulse_period).to(unit=unit, copy=False)
        )
    return UnwrappedTimeOfArrival(toa)


def unwrapped_time_of_arrival_minus_frame_pivot_time(
    toa: UnwrappedTimeOfArrival, pivot_time: PivotTimeAtDetector
) -> UnwrappedTimeOfArrivalMinusPivotTime:
    """
    Compute the time of arrival of the neutron at the detector, unwrapped at the pulse
    period, minus the start time of the frame.
    We subtract the start time of the frame so that we can use a modulo operation to
    wrap the time of arrival at the frame period in the case of pulse-skipping.

    Parameters
    ----------
    toa:
        Time of arrival of the neutron at the detector, unwrapped at the pulse period.
    pivot_time:
        Pivot time at the detector, i.e., the time of the start of the frame at the
        detector.
    """
    # Order of operation to preserve dimension order
    return UnwrappedTimeOfArrivalMinusPivotTime(
        -pivot_time.to(unit=elem_unit(toa), copy=False) + toa
    )


def time_of_arrival_minus_pivot_time_modulo_period(
    toa_minus_pivot_time: UnwrappedTimeOfArrivalMinusPivotTime,
    frame_period: FramePeriod,
) -> TimeOfArrivalMinusPivotTimeModuloPeriod:
    """
    Compute the time of arrival of the neutron at the detector, unwrapped at the pulse
    period, minus the start time of the frame, modulo the frame period.

    Parameters
    ----------
    toa_minus_pivot_time:
        Time of arrival of the neutron at the detector, unwrapped at the pulse period,
        minus the start time of the frame.
    frame_period:
        Period of the frame, i.e., time between the start of two consecutive frames.
    """
    return TimeOfArrivalMinusPivotTimeModuloPeriod(
        toa_minus_pivot_time
        % frame_period.to(unit=elem_unit(toa_minus_pivot_time), copy=False)
    )


def time_of_arrival_folded_by_frame(
    toa: TimeOfArrivalMinusPivotTimeModuloPeriod,
    pivot_time: PivotTimeAtDetector,
) -> FrameFoldedTimeOfArrival:
    """
    The time of arrival of the neutron at the detector, folded by the frame period.

    Parameters
    ----------
    toa:
        Time of arrival of the neutron at the detector, unwrapped at the pulse period,
        minus the start time of the frame, modulo the frame period.
    pivot_time:
        Pivot time at the detector, i.e., the time of the start of the frame at the
        detector.
    """
    return FrameFoldedTimeOfArrival(
        toa + pivot_time.to(unit=elem_unit(toa), copy=False)
    )


def time_of_flight_data(
    da: RawData,
    lookup: MaskedTimeOfFlightLookupTable,
    ltotal: Ltotal,
    toas: FrameFoldedTimeOfArrival,
) -> TofData:
    """
    Convert the time-of-arrival data to time-of-flight data using a lookup table.
    The output data will have a time-of-flight coordinate.

    Parameters
    ----------
    da:
        Raw detector data loaded from a NeXus file, e.g., NXdetector containing
        NXevent_data.
    lookup:
        Lookup table giving time-of-flight as a function of distance and time of
        arrival.
    ltotal:
        Total length of the flight path from the source to the detector.
    toas:
        Time of arrival of the neutron at the detector, folded by the frame period.
    """
    from scipy.interpolate import RegularGridInterpolator

    # TODO: to make use of multi-threading, we could write our own interpolator.
    # This should be simple enough as we are making the bins linspace, so computing
    # bin indices is fast.
    f = RegularGridInterpolator(
        (
            lookup.coords["toa"].to(unit=elem_unit(toas), copy=False).values,
            lookup.coords["distance"].to(unit=ltotal.unit, copy=False).values,
        ),
        lookup.data.to(unit=elem_unit(toas), copy=False).values.T,
        method="linear",
        bounds_error=False,
    )

    if da.bins is not None:
        ltotal = sc.bins_like(toas, ltotal).bins.constituents["data"]
        toas = toas.bins.constituents["data"]

    tofs = sc.array(
        dims=toas.dims, values=f((toas.values, ltotal.values)), unit=elem_unit(toas)
    )

    if da.bins is not None:
        parts = da.bins.constituents
        parts["data"] = tofs
        out = da.bins.assign_coords(tof=_bins_no_validate(**parts))
    else:
        out = da.assign_coords(tof=tofs)

    return TofData(out)


def resample_tof_data(da: TofData) -> ResampledTofData:
    """
    Histogrammed data that has been converted to `tof` will typically have
    unsorted bin edges (due to either wrapping of `time_of_flight` or wavelength
    overlap between subframes).
    This function re-histograms the data to ensure that the bin edges are sorted.
    It makes use of the ``to_events`` helper which generates a number of events in each
    bin with a uniform distribution. The new events are then histogrammed using a set of
    sorted bin edges.

    WARNING:
    This function is highly experimental, has limitations and should be used with
    caution. It is a workaround to the issue that rebinning data with unsorted bin
    edges is not supported in scipp.
    As such, this function is not part of the default set of providers, and needs to be
    inserted manually into the workflow.

    Parameters
    ----------
    da:
        Histogrammed data with the time-of-flight coordinate.
    """
    dim = next(iter(set(da.dims) & {"time_of_flight", "tof"}))
    events = to_events(da.rename_dims({dim: "tof"}), "event")

    # Define a new bin width, close to the original bin width.
    # TODO: this could be a workflow parameter
    coord = da.coords["tof"]
    bin_width = (coord[dim, 1:] - coord[dim, :-1]).nanmedian()
    rehist = events.hist(tof=bin_width)
    for key, var in da.coords.items():
        if dim not in var.dims:
            rehist.coords[key] = var
    return ResampledTofData(rehist)


def default_parameters() -> dict:
    """
    Default parameters of the time-of-flight workflow.
    """
    return {
        PulsePeriod: 1.0 / sc.scalar(14.0, unit="Hz"),
        PulseStride: 1,
        PulseStrideOffset: 0,
        DistanceResolution: sc.scalar(0.1, unit="m"),
        TimeOfArrivalResolution: 500,
        LookupTableRelativeErrorThreshold: 0.1,
    }


def providers() -> tuple[Callable]:
    """
    Providers of the time-of-flight workflow.
    """
    return (
        compute_tof_lookup_table,
        extract_ltotal,
        find_fastest_neutron,
        frame_period,
        masked_tof_lookup_table,
        pivot_time_at_detector,
        time_of_arrival_folded_by_frame,
        time_of_arrival_minus_pivot_time_modulo_period,
        time_of_flight_data,
        unwrapped_time_of_arrival,
        unwrapped_time_of_arrival_minus_frame_pivot_time,
    )


class TofWorkflow:
    """
    Helper class to build a time-of-flight workflow and cache the expensive part of the
    computation: running the simulation and building the lookup table.

    Parameters
    ----------
    simulated_neutrons:
        Results of a time-of-flight simulation used to create a lookup table.
        The results should be a flat table with columns for time-of-arrival, speed,
        wavelength, and weight.
    ltotal_range:
        Range of total flight path lengths from the source to the detector.
        This is used to create the lookup table to compute the neutron
        time-of-flight.
        Note that the resulting table will extend slightly beyond this range, as the
        supplied range is not necessarily a multiple of the distance resolution.
    pulse_stride:
        Stride of used pulses. Usually 1, but may be a small integer when
        pulse-skipping.
    pulse_stride_offset:
        Integer offset of the first pulse in the stride (typically zero unless we
        are using pulse-skipping and the events do not begin with the first pulse in
        the stride).
    distance_resolution:
        Resolution of the distance axis in the lookup table.
        Should be a single scalar value with a unit of length.
        This is typically of the order of 1-10 cm.
    toa_resolution:
        Resolution of the time of arrival axis in the lookup table.
        Can be an integer (number of bins) or a sc.Variable (bin width).
    error_threshold:
        Threshold for the variance of the projected time-of-flight above which
        regions are masked.
    """

    def __init__(
        self,
        simulated_neutrons: SimulationResults,
        ltotal_range: LtotalRange,
        pulse_stride: PulseStride | None = None,
        pulse_stride_offset: PulseStrideOffset | None = None,
        distance_resolution: DistanceResolution | None = None,
        toa_resolution: TimeOfArrivalResolution | None = None,
        error_threshold: LookupTableRelativeErrorThreshold | None = None,
    ):
        import sciline as sl

        self.pipeline = sl.Pipeline(providers())
        self.pipeline[SimulationResults] = simulated_neutrons
        self.pipeline[LtotalRange] = ltotal_range

        params = default_parameters()
        self.pipeline[PulsePeriod] = params[PulsePeriod]
        self.pipeline[PulseStride] = pulse_stride or params[PulseStride]
        self.pipeline[PulseStrideOffset] = (
            pulse_stride_offset or params[PulseStrideOffset]
        )
        self.pipeline[DistanceResolution] = (
            distance_resolution or params[DistanceResolution]
        )
        self.pipeline[TimeOfArrivalResolution] = (
            toa_resolution or params[TimeOfArrivalResolution]
        )
        self.pipeline[LookupTableRelativeErrorThreshold] = (
            error_threshold or params[LookupTableRelativeErrorThreshold]
        )

    def cache_results(
        self,
        results=(SimulationResults, MaskedTimeOfFlightLookupTable, FastestNeutron),
    ) -> None:
        """
        Cache a list of (usually expensive to compute) intermediate results of the
        time-of-flight workflow.

        Parameters
        ----------
        results:
            List of results to cache.
        """
        for t in results:
            self.pipeline[t] = self.pipeline.compute(t)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
"""NeXus utilities.

This module defines functions and domain types that can be used
to build Sciline pipelines for simple workflows.
If multiple different kinds of files (e.g., sample and background runs)
are needed, custom types and providers need to be defined to wrap
the basic ones here.

Providers / functions are available from here directly.
The submodule :mod:`types` defines all domain types.
"""

from . import types
from .workflow import GenericNeXusWorkflow
from ._nexus_loader import (
    load_data,
    group_event_data,
    load_component,
    load_all_components,
    compute_component_position,
    extract_signal_data_array,
)

__all__ = [
    'types',
    'group_event_data',
    'load_all_components',
    'load_data',
    'load_component',
    'compute_component_position',
    'extract_signal_data_array',
    'GenericNeXusWorkflow',
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical
from scippneutron.conversion.graph.beamline import beamline as beamline_graph
from scippneutron.conversion.graph.tof import elastic as elastic_graph

from ess.reduce import time_of_flight
from ess.reduce.time_of_flight import fakes

sl = pytest.importorskip("sciline")


def test_frame_period_is_pulse_period_if_not_pulse_skipping() -> None:
    pl = sl.Pipeline(time_of_flight.providers())
    period = sc.scalar(123.0, unit="ms")
    pl[time_of_flight.PulsePeriod] = period
    pl[time_of_flight.PulseStride] = 1
    assert_identical(pl.compute(time_of_flight.FramePeriod), period)


@pytest.mark.parametrize("stride", [1, 2, 3, 4])
def test_frame_period_is_multiple_pulse_period_if_pulse_skipping(stride) -> None:
    pl = sl.Pipeline(time_of_flight.providers())
    period = sc.scalar(123.0, unit="ms")
    pl[time_of_flight.PulsePeriod] = period
    pl[time_of_flight.PulseStride] = stride
    assert_identical(pl.compute(time_of_flight.FramePeriod), stride * period)


def test_unwrap_with_no_choppers() -> None:
    # At this small distance the frames are not overlapping (with the given wavelength
    # range), despite not using any choppers.
    distance = sc.scalar(10.0, unit="m")

    beamline = fakes.FakeBeamline(
        choppers={},
        monitors={"detector": distance},
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=100_000,
    )

    mon, ref = beamline.get_monitor("detector")

    pl = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    sim = time_of_flight.simulate_beamline(choppers={}, neutrons=300_000, seed=1234)

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    pl[time_of_flight.LookupTableRelativeErrorThreshold] = 1.0

    tofs = pl.compute(time_of_flight.TofData)

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    ref = ref.bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # Most errors should be small
    assert np.nanpercentile(diff.values, 96) < 1.0


# At 80m, event_time_offset does not wrap around (all events are within the same pulse).
# At 85m, event_time_offset wraps around.
@pytest.mark.parametrize("dist", [80.0, 85.0])
def test_standard_unwrap(dist) -> None:
    distance = sc.scalar(dist, unit="m")
    choppers = fakes.psc_choppers()
    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance

    tofs = pl.compute(time_of_flight.TofData)

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    ref = ref.bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # All errors should be small
    assert np.nanpercentile(diff.values, 100) < 0.01


# At 80m, event_time_offset does not wrap around (all events are within the same pulse).
# At 85m, event_time_offset wraps around.
@pytest.mark.parametrize("dist", [80.0, 85.0])
@pytest.mark.parametrize("dim", ["time_of_flight", "tof"])
def test_standard_unwrap_histogram_mode(dist, dim) -> None:
    distance = sc.scalar(dist, unit="m")
    choppers = fakes.psc_choppers()
    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1 / 14, unit="s") * 4,
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")
    mon = (
        mon.hist(
            event_time_offset=sc.linspace(
                "event_time_offset", 0.0, 1000.0 / 14, num=1001, unit="ms"
            ).to(unit="s")
        )
        .sum("pulse")
        .rename(event_time_offset=dim)
    )

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        (*time_of_flight.providers(), time_of_flight.resample_tof_data),
        params=time_of_flight.default_parameters(),
    )

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    tofs = pl.compute(time_of_flight.ResampledTofData)
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)
    ref = ref.bins.concat().value.hist(wavelength=wavs.coords["wavelength"])
    # We divide by the maximum to avoid large relative differences at the edges of the
    # frames where the counts are low.
    diff = (wavs - ref) / ref.max()
    assert np.nanpercentile(diff.values, 96.0) < 0.3


def test_pulse_skipping_unwrap() -> None:
    distance = sc.scalar(100.0, unit="m")
    choppers = fakes.psc_choppers()
    choppers["pulse_skipping"] = fakes.pulse_skipping

    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1.0, unit="s"),
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    pl[time_of_flight.PulseStride] = 2

    tofs = pl.compute(time_of_flight.TofData)

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    ref = ref.bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # All errors should be small
    assert np.nanpercentile(diff.values, 100) < 0.01


def test_pulse_skipping_unwrap_when_all_neutrons_arrive_after_second_pulse() -> None:
    distance = sc.scalar(150.0, unit="m")
    choppers = fakes.psc_choppers()
    choppers["pulse_skipping"] = fakes.pulse_skipping

    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1.0, unit="s"),
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    pl[time_of_flight.PulseStride] = 2
    pl[time_of_flight.PulseStrideOffset] = 1  # Start the stride at the second pulse

    tofs = pl.compute(time_of_flight.TofData)

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    ref = ref.bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # All errors should be small
    assert np.nanpercentile(diff.values, 100) < 0.01


def test_pulse_skipping_unwrap_when_first_half_of_first_pulse_is_missing() -> None:
    distance = sc.scalar(100.0, unit="m")
    choppers = fakes.psc_choppers()
    choppers["pulse_skipping"] = fakes.pulse_skipping

    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1.0, unit="s"),
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    pl[time_of_flight.RawData] = mon[
        1:
    ].copy()  # Skip first pulse = half of the first frame
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    pl[time_of_flight.PulseStride] = 2
    pl[time_of_flight.PulseStrideOffset] = 1  # Start the stride at the second pulse

    tofs = pl.compute(time_of_flight.TofData)

    # Convert to wavelength
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph).bins.concat().value
    ref = ref[1:].copy().bins.concat().value

    diff = abs(
        (wavs.coords["wavelength"] - ref.coords["wavelength"])
        / ref.coords["wavelength"]
    )
    # All errors should be small
    assert np.nanpercentile(diff.values, 100) < 0.01


def test_pulse_skipping_unwrap_histogram_mode() -> None:
    distance = sc.scalar(100.0, unit="m")
    choppers = fakes.psc_choppers()
    choppers["pulse_skipping"] = fakes.pulse_skipping

    beamline = fakes.FakeBeamline(
        choppers=choppers,
        monitors={"detector": distance},
        run_length=sc.scalar(1.0, unit="s"),
        events_per_pulse=100_000,
    )
    mon, ref = beamline.get_monitor("detector")
    mon = (
        mon.hist(
            event_time_offset=sc.linspace(
                "event_time_offset", 0.0, 1000.0 / 14, num=1001, unit="ms"
            ).to(unit="s")
        )
        .sum("pulse")
        .rename(event_time_offset="time_of_flight")
    )

    sim = time_of_flight.simulate_beamline(
        choppers=choppers, neutrons=300_000, seed=1234
    )

    pl = sl.Pipeline(
        (*time_of_flight.providers(), time_of_flight.resample_tof_data),
        params=time_of_flight.default_parameters(),
    )

    pl[time_of_flight.RawData] = mon
    pl[time_of_flight.SimulationResults] = sim
    pl[time_of_flight.LtotalRange] = distance, distance
    pl[time_of_flight.PulseStride] = 2
    tofs = pl.compute(time_of_flight.ResampledTofData)
    graph = {**beamline_graph(scatter=False), **elastic_graph("tof")}
    wavs = tofs.transform_coords("wavelength", graph=graph)
    ref = ref.bins.concat().value.hist(wavelength=wavs.coords["wavelength"])
    # We divide by the maximum to avoid large relative differences at the edges of the
    # frames where the counts are low.
    diff = (wavs - ref) / ref.max()
    assert np.nanpercentile(diff.values, 96.0) < 0.3



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
"""
A fake time-of-flight neutron beamline for documentation and testing.

This provides detector event data in a structure as typically provided in a NeXus file,
with event_time_offset and event_time_zero information.
"""

from collections.abc import Callable

import numpy as np
import scipp as sc
from scippneutron.chopper import DiskChopper


class FakeBeamline:
    def __init__(
        self,
        choppers: dict[str, DiskChopper],
        monitors: dict[str, sc.Variable],
        run_length: sc.Variable,
        events_per_pulse: int = 200000,
        source: Callable | None = None,
    ):
        import math

        import tof as tof_pkg
        from tof.facilities.ess_pulse import pulse

        self.frequency = pulse.frequency
        self.npulses = math.ceil((run_length * self.frequency).to(unit="").value)
        self.events_per_pulse = events_per_pulse

        # Create a source
        if source is None:
            self.source = tof_pkg.Source(
                facility="ess", neutrons=self.events_per_pulse, pulses=self.npulses
            )
        else:
            self.source = source(pulses=self.npulses)

        # Convert the choppers to tof.Chopper
        self.choppers = [
            tof_pkg.Chopper(
                frequency=abs(ch.frequency),
                direction=tof_pkg.AntiClockwise
                if (ch.frequency.value > 0.0)
                else tof_pkg.Clockwise,
                open=ch.slit_begin,
                close=ch.slit_end,
                phase=abs(ch.phase),
                distance=ch.axle_position.fields.z,
                name=name,
            )
            for name, ch in choppers.items()
        ]

        # Add detectors
        self.monitors = [
            tof_pkg.Detector(distance=distance, name=key)
            for key, distance in monitors.items()
        ]

        #  Propagate the neutrons
        self.model = tof_pkg.Model(
            source=self.source, choppers=self.choppers, detectors=self.monitors
        )
        self.model_result = self.model.run()

    def get_monitor(self, name: str) -> sc.DataGroup:
        # Create some fake pulse time zero
        start = sc.datetime("2024-01-01T12:00:00.000000")
        period = sc.reciprocal(self.frequency)

        detector = self.model_result.detectors[name]
        raw_data = detector.data.flatten(to="event")
        # Select only the neutrons that make it to the detector
        raw_data = raw_data[~raw_data.masks["blocked_by_others"]].copy()
        raw_data.coords["Ltotal"] = detector.distance

        # Format the data in a way that resembles data loaded from NeXus
        event_data = raw_data.copy(deep=False)
        dt = period.to(unit="us")
        event_time_zero = (dt * (event_data.coords["toa"] // dt)).to(dtype=int) + start
        raw_data.coords["event_time_zero"] = event_time_zero
        event_data.coords["event_time_zero"] = event_time_zero
        event_data.coords["event_time_offset"] = (
            event_data.coords.pop("toa").to(unit="s") % period
        )
        del event_data.coords["tof"]
        del event_data.coords["speed"]
        del event_data.coords["time"]
        del event_data.coords["wavelength"]

        return (
            event_data.group("event_time_zero").rename_dims(event_time_zero="pulse"),
            raw_data.group("event_time_zero").rename_dims(event_time_zero="pulse"),
        )


wfm1_chopper = DiskChopper(
    frequency=sc.scalar(-70.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(-47.10, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 6.6], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([83.71, 140.49, 193.26, 242.32, 287.91, 330.3]) + 15.0,
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([94.7, 155.79, 212.56, 265.33, 314.37, 360.0]) + 15.0,
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)

wfm2_chopper = DiskChopper(
    frequency=sc.scalar(-70.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(-76.76, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 7.1], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([65.04, 126.1, 182.88, 235.67, 284.73, 330.32]) + 15.0,
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([76.03, 141.4, 202.18, 254.97, 307.74, 360.0]) + 15.0,
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)

foc1_chopper = DiskChopper(
    frequency=sc.scalar(-56.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(-62.40, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 8.8], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([74.6, 139.6, 194.3, 245.3, 294.8, 347.2]),
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([95.2, 162.8, 216.1, 263.1, 310.5, 371.6]),
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)

foc2_chopper = DiskChopper(
    frequency=sc.scalar(-28.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(-12.27, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 15.9], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([98.0, 154.0, 206.8, 255.0, 299.0, 344.65]),
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([134.6, 190.06, 237.01, 280.88, 323.56, 373.76]),
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)

pol_chopper = DiskChopper(
    frequency=sc.scalar(-14.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(0.0, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 17.0], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([40.0]),
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([240.0]),
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)

pulse_skipping = DiskChopper(
    frequency=sc.scalar(-7.0, unit="Hz"),
    beam_position=sc.scalar(0.0, unit="deg"),
    phase=sc.scalar(0.0, unit="deg"),
    axle_position=sc.vector(value=[0, 0, 30.0], unit="m"),
    slit_begin=sc.array(
        dims=["cutout"],
        values=np.array([40.0]),
        unit="deg",
    ),
    slit_end=sc.array(
        dims=["cutout"],
        values=np.array([140.0]),
        unit="deg",
    ),
    slit_height=sc.scalar(10.0, unit="cm"),
    radius=sc.scalar(30.0, unit="cm"),
)


def wfm_choppers():
    return {
        "wfm1": wfm1_chopper,
        "wfm2": wfm2_chopper,
        "foc1": foc1_chopper,
        "foc2": foc2_chopper,
        "pol": pol_chopper,
    }


def psc_choppers():
    return {
        name: DiskChopper(
            frequency=ch.frequency,
            beam_position=ch.beam_position,
            phase=ch.phase,
            axle_position=ch.axle_position,
            slit_begin=ch.slit_begin[0:1],
            slit_end=ch.slit_end[0:1],
            slit_height=ch.slit_height[0:1],
            radius=ch.radius,
        )
        for name, ch in wfm_choppers().items()
    }



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_identical

from ess.reduce import uncertainty as unc


def test_fail_mode_always_returns_original():
    broadcaster = unc.broadcasters[unc.UncertaintyBroadcastMode.fail]
    var = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    assert broadcaster(var, prototype=sc.zeros(sizes={'x': 2})) is var
    assert broadcaster(var, prototype=sc.zeros(sizes={'y': 3})) is var
    assert broadcaster(var, prototype=sc.zeros(sizes={'x': 2, 'y': 3})) is var
    assert broadcaster(var, prototype=sc.zeros(sizes={'z': 5})) is var


@pytest.mark.parametrize(
    'mode',
    [unc.UncertaintyBroadcastMode.drop, unc.UncertaintyBroadcastMode.upper_bound],
)
def test_broadcaster_returns_original_if_no_new_dims(mode):
    broadcaster = unc.broadcasters[mode]
    var = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    assert broadcaster(var, prototype=sc.zeros(sizes={'x': 2})) is var
    assert broadcaster(var, prototype=sc.zeros(sizes={'y': 3})) is var
    assert broadcaster(var, prototype=sc.zeros(sizes={'x': 2, 'y': 3})) is var


@pytest.mark.parametrize(
    'mode',
    [unc.UncertaintyBroadcastMode.drop, unc.UncertaintyBroadcastMode.upper_bound],
)
def test_broadcaster_original_if_no_variances(mode):
    broadcaster = unc.broadcasters[mode]
    var = sc.ones(dims=['x'], shape=[2], with_variances=False)
    assert broadcaster(var, prototype=sc.zeros(sizes={'y': 3})) is var


def test_broadcast_scales_variances_by_new_subspace_volume():
    x = sc.linspace('x', 0.0, 1.0, 2)
    y = sc.linspace('y', 0.0, 2.0, 3)
    values = x * y
    var = values.copy()
    var.variances = var.values
    expected = sc.ones(dims=['z'], shape=[1]) * values
    expected.variances = 1 * expected.values
    assert_identical(
        unc.broadcast_with_upper_bound_variances(
            var, prototype=sc.zeros(sizes={'z': 1})
        ),
        expected,
    )
    expected = sc.ones(dims=['z'], shape=[2]) * values
    expected.variances = 2 * expected.values
    assert_identical(
        unc.broadcast_with_upper_bound_variances(
            var, prototype=sc.zeros(sizes={'z': 2})
        ),
        expected,
    )
    expected = sc.ones(dims=['z'], shape=[2]) * values
    expected.variances = 2 * expected.values
    assert_identical(
        unc.broadcast_with_upper_bound_variances(
            var, prototype=sc.zeros(sizes={'y': 3, 'z': 2})
        ),
        expected.transpose(['y', 'z', 'x']),
    )


def test_drop_mode_drops_if_new_dims():
    broadcaster = unc.broadcasters[unc.UncertaintyBroadcastMode.drop]
    var = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    assert_identical(
        broadcaster(var, prototype=sc.zeros(sizes={'z': 1})), sc.values(var)
    )


def test_broadcast_with_mask_along_single_broadcast_dim_does_not_count_masked():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    y = sc.DataArray(
        sc.ones(dims=['y'], shape=[3]),
        masks={'y': sc.array(dims=['y'], values=[False, True, False])},
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['y', 'x'], shape=[3, 2], with_variances=True)
    expected.variances *= 2
    expected['y', 1].variances = [np.inf, np.inf]
    assert_identical(xy, expected)


def test_broadcast_with_mask_along_existing_dim_has_no_effect():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    y = sc.DataArray(
        sc.ones(dims=['y'], shape=[3]),
        masks={'x': sc.array(dims=['x'], values=[True, False])},
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['y', 'x'], shape=[3, 2], with_variances=True)
    expected.variances *= 3
    assert_identical(xy, expected)


def test_broadcast_2d_with_mask_along_single_broadcast_dim_does_not_count_masked():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    y = sc.DataArray(
        sc.ones(dims=['z', 'y'], shape=[2, 3]),
        masks={'y': sc.array(dims=['y'], values=[False, True, False])},
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['z', 'y', 'x'], shape=[2, 3, 2], with_variances=True)
    expected.variances *= 4
    expected['y', 1].variances = [[np.inf, np.inf], [np.inf, np.inf]]
    assert_identical(xy, expected)


def test_broadcast_into_orthogonal_2d_mask_does_not_count_masked():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    y = sc.DataArray(
        sc.ones(dims=['z', 'y'], shape=[2, 3]),
        masks={
            'zy': sc.array(
                dims=['z', 'y'], values=[[False, False, False], [False, True, False]]
            )
        },
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['z', 'y', 'x'], shape=[2, 3, 2], with_variances=True)
    expected.variances *= 5
    expected['z', 1]['y', 1].variances = [np.inf, np.inf]
    assert_identical(xy, expected)


def test_broadcast_into_nonorthogonal_2d_mask_ignores_mask():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    # There is no simple way of handling this, so we just ignore the mask.
    # This is ok since it will simply mean the upper bound is larger than strictly
    # necessary.
    y = sc.DataArray(
        sc.ones(dims=['x', 'y'], shape=[2, 3]),
        masks={
            'xy': sc.array(
                dims=['x', 'y'], values=[[False, False, False], [False, True, False]]
            )
        },
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    expected.variances *= 3
    assert_identical(xy, expected)


def test_broadcast_into_nonorthogonal_2d_mask_reducible_mask_counts_masked():
    x = sc.ones(dims=['x'], shape=[2], with_variances=True)
    # There is no simple way of handling this, so we just ignore the mask.
    # This is ok since it will simply mean the upper bound is larger than strictly
    # necessary.
    y = sc.DataArray(
        sc.ones(dims=['x', 'y'], shape=[2, 3]),
        masks={
            'xy': sc.array(
                dims=['x', 'y'], values=[[False, True, False], [False, True, False]]
            )
        },
    )
    xy = unc.broadcast_with_upper_bound_variances(x, prototype=y)
    expected = sc.ones(dims=['x', 'y'], shape=[2, 3], with_variances=True)
    expected.variances *= 2
    expected['y', 1].variances = [np.inf, np.inf]
    assert_identical(xy, expected)


def test_upper_bound_broadcast_raises_if_input_is_binned():
    x = sc.linspace('x', 0.0, 1.0, 10).bin(x=1).squeeze()
    x.value.variances = x.value.values
    y = sc.linspace('y', 0.0, 1.0, 10)
    with pytest.raises(ValueError, match="Cannot broadcast binned data."):
        unc.broadcast_with_upper_bound_variances(x, prototype=y)


def test_upper_bound_event_broadcast_raises_if_binning_mismatching():
    prototype = sc.linspace('x', 0.0, 1.0, 10).bin(x=3).squeeze()
    data = sc.DataArray(
        sc.ones(dims=['x'], shape=[2], with_variances=True),
        coords={'x': sc.linspace('x', 0.0, 1.0, 3)},
    )
    with pytest.raises(
        ValueError, match="Mismatching binning not supported in broadcast."
    ):
        unc.broadcast_with_upper_bound_variances(data, prototype=prototype)


def test_upper_bound_event_broadcast_counts_events():
    content = sc.ones(dims=['event'], shape=[10])
    # sizes = [0,1,2,4,3]
    begin = sc.array(dims=['x'], values=[0, 0, 1, 3, 7], unit=None)
    prototype = sc.bins(data=content, dim='event', begin=begin)
    # We are not supporting the case of prototype missing dims of the data.
    prototype = sc.concat([prototype, prototype], 'y')
    y = sc.array(dims=['y'], values=[1.0, 2.0], variances=[1.0, 2.0])
    upper_bound_broadcast = unc.broadcast_with_upper_bound_variances(
        y, prototype=prototype
    )

    expected = prototype.copy().bins.constituents
    expected['data'].values[:10] = 1.0
    expected['data'].values[10:] = 2.0
    # There are 5 bins along x, but 10 events, so variance scale factor is 10.
    expected['data'].variances = expected['data'].values * 10
    expected = sc.bins(**expected)

    assert_identical(upper_bound_broadcast, expected)
    # The point of broadcast_with_upper_bound_variances is that we can afterwards
    # perform the following operation with getting a variance broadcast error.
    # Did it work?
    _ = prototype * upper_bound_broadcast


def test_upper_bound_event_broadcast_event_count_excludes_masked():
    content = sc.ones(dims=['event'], shape=[10])
    # sizes = [0,1,2,4,3]
    begin = sc.array(dims=['x'], values=[0, 0, 1, 3, 7], unit=None)
    prototype = sc.bins(data=content, dim='event', begin=begin)
    # We are not supporting the case of prototype missing dims of the data.
    prototype = sc.DataArray(sc.concat([prototype, prototype], 'y'))
    prototype.masks['x'] = sc.array(
        dims=['x'], values=[False, True, False, False, True]
    )
    y = sc.array(dims=['y'], values=[1.0, 2.0], variances=[1.0, 2.0])
    upper_bound_broadcast = unc.broadcast_with_upper_bound_variances(
        y, prototype=prototype
    )

    expected = prototype.copy().bins.constituents
    expected['data'].values[:10] = 1.0
    expected['data'].values[10:] = 2.0
    # There are 5 bins along x, but 10 events (4 masked), so variance scale factor is 6.
    expected['data'].variances = expected['data'].values * 6
    expected['data']['event', 0:1].variances = [np.inf]
    expected['data']['event', 7:10].variances = [np.inf, np.inf, np.inf]
    expected['data']['event', 10:11].variances = [np.inf]
    expected['data']['event', 17:20].variances = [np.inf, np.inf, np.inf]
    expected = sc.bins(**expected)

    assert_identical(upper_bound_broadcast, expected)
    # The point of broadcast_with_upper_bound_variances is that we can afterwards
    # perform the following operation with getting a variance broadcast error.
    # Did it work?
    _ = prototype * upper_bound_broadcast
    assert sc.all(sc.isinf(sc.variances(upper_bound_broadcast['x', 1])))
    assert sc.all(sc.isinf(sc.variances(upper_bound_broadcast['x', 4])))


def test_upper_bound_event_broadcast_only_bin_broadcast():
    content = sc.ones(dims=['event'], shape=[10])
    begin = sc.array(dims=['x'], values=[0, 3, 6], unit=None)
    prototype = sc.DataArray(sc.bins(data=content, dim='event', begin=begin))
    prototype.masks['x'] = sc.array(dims=['x'], values=[False, True, False])
    x = sc.array(dims=['x'], values=[1.0, 2.0, 3.0], variances=[1.0, 2.0, 3.0])
    # Data has same dims as prototype, broadcast is only into bins
    assert x.sizes == prototype.sizes
    upper_bound_broadcast = unc.broadcast_with_upper_bound_variances(
        x, prototype=prototype
    )

    expected = prototype.copy().bins.constituents
    expected['data'].values[:3] = 1.0
    expected['data'].values[3:6] = 2.0
    expected['data'].values[6:] = 3.0
    expected['data'].variances = expected['data'].values
    expected['data']['event', 0:3].variances = [3.0, 3.0, 3.0]
    # The x=1 bin is masked, but it is not being broadcasted, so no inf.
    expected['data']['event', 3:6].variances = [6.0, 6.0, 6.0]
    expected['data']['event', 6:].variances = [12.0, 12.0, 12.0, 12.0]
    expected = sc.bins(**expected)

    assert_identical(upper_bound_broadcast, expected)
    # The point of broadcast_with_upper_bound_variances is that we can afterwards
    # perform the following operation with getting a variance broadcast error.
    # Did it work?
    _ = prototype * upper_bound_broadcast



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from functools import partial

import pytest
import scipp as sc
import tof as tof_pkg
from scippneutron.chopper import DiskChopper
from scippneutron.conversion.graph.beamline import beamline
from scippneutron.conversion.graph.tof import elastic

from ess.reduce import time_of_flight
from ess.reduce.time_of_flight import fakes

sl = pytest.importorskip("sciline")


def dream_choppers():
    psc1 = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(286 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.145], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 70.49, 84.765, 113.565, 170.29, 271.635, 286.035, 301.17],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 73.51, 88.035, 116.835, 175.31, 275.565, 289.965, 303.63],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    psc2 = DiskChopper(
        frequency=sc.scalar(-14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(-236, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.155], unit="m"),
        slit_begin=sc.array(
            dims=["cutout"],
            values=[-1.23, 27.0, 55.8, 142.385, 156.765, 214.115, 257.23, 315.49],
            unit="deg",
        ),
        slit_end=sc.array(
            dims=["cutout"],
            values=[1.23, 30.6, 59.4, 145.615, 160.035, 217.885, 261.17, 318.11],
            unit="deg",
        ),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    oc = DiskChopper(
        frequency=sc.scalar(14.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(297 - 180 - 90, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 6.174], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-27.6 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[27.6 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    bcc = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 9.78], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[36.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    t0 = DiskChopper(
        frequency=sc.scalar(28.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(280 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 13.05], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-314.9 * 0.5], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[314.9 * 0.5], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )

    return {"psc1": psc1, "psc2": psc2, "oc": oc, "bcc": bcc, "t0": t0}


def dream_choppers_with_frame_overlap():
    out = dream_choppers()
    out["bcc"] = DiskChopper(
        frequency=sc.scalar(112.0, unit="Hz"),
        beam_position=sc.scalar(0.0, unit="deg"),
        phase=sc.scalar(240 - 180, unit="deg"),
        axle_position=sc.vector(value=[0, 0, 9.78], unit="m"),
        slit_begin=sc.array(dims=["cutout"], values=[-36.875, 143.125], unit="deg"),
        slit_end=sc.array(dims=["cutout"], values=[56.875, 216.875], unit="deg"),
        slit_height=sc.scalar(10.0, unit="cm"),
        radius=sc.scalar(30.0, unit="cm"),
    )
    return out


@pytest.fixture(scope="module")
def simulation_dream_choppers():
    return time_of_flight.simulate_beamline(
        choppers=dream_choppers(), neutrons=100_000, seed=432
    )


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[77.675], unit="m"),
        sc.array(dims=["detector_number"], values=[77.675, 76.5], unit="m"),
        sc.array(
            dims=["y", "x"],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit="m",
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_dream_wfm(
    simulation_dream_choppers,
    npulses,
    ltotal,
    time_offset_unit,
    distance_unit,
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    wavelengths = sc.array(
        dims=["event"], values=[1.5, 1.6, 1.7, 3.3, 3.4, 3.5], unit="angstrom"
    )
    birth_times = sc.full(sizes=wavelengths.sizes, value=1.5, unit="ms")
    ess_beamline = fakes.FakeBeamline(
        choppers=dream_choppers(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key)[0] for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw_data.bins.coords["event_time_offset"] = raw_data.bins.coords[
        "event_time_offset"
    ].to(unit=time_offset_unit, copy=False)

    raw_data.coords["Ltotal"] = ltotal.to(unit=distance_unit, copy=False)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat("pulse").hist().data,
        sc.array(
            dims=["detector"],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype="float64",
        ).fold(dim="detector", sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    workflow[time_of_flight.RawData] = raw_data
    workflow[time_of_flight.SimulationResults] = simulation_dream_choppers
    workflow[time_of_flight.LtotalRange] = ltotal.min(), ltotal.max()

    # Compute time-of-flight
    tofs = workflow.compute(time_of_flight.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result = wav_wfm["pulse", i].flatten(to="detector")
        for j in range(len(result)):
            computed_wavelengths = result[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths,
                true_wavelengths["pulse", i],
                rtol=sc.scalar(1e-02),
            )


@pytest.fixture(scope="module")
def simulation_dream_choppers_time_overlap():
    return time_of_flight.simulate_beamline(
        choppers=dream_choppers_with_frame_overlap(), neutrons=100_000, seed=432
    )


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[77.675], unit="m"),
        sc.array(dims=["detector_number"], values=[77.675, 76.5], unit="m"),
        sc.array(
            dims=["y", "x"],
            values=[[77.675, 76.1, 78.05], [77.15, 77.3, 77.675]],
            unit="m",
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_dream_wfm_with_subframe_time_overlap(
    simulation_dream_choppers_time_overlap,
    npulses,
    ltotal,
    time_offset_unit,
    distance_unit,
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    wavelengths = [1.5, 1.6, 1.7, 3.3, 3.4, 3.5]
    birth_times = [1.5] * len(wavelengths)

    # Add overlap neutrons
    birth_times.extend([0.0, 3.3])
    wavelengths.extend([2.6, 2.4])

    wavelengths = sc.array(dims=["event"], values=wavelengths, unit="angstrom")
    birth_times = sc.array(dims=["event"], values=birth_times, unit="ms")

    ess_beamline = fakes.FakeBeamline(
        choppers=dream_choppers_with_frame_overlap(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key)[0] for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw_data.bins.coords["event_time_offset"] = raw_data.bins.coords[
        "event_time_offset"
    ].to(unit=time_offset_unit, copy=False)

    raw_data.coords["Ltotal"] = ltotal.to(unit=distance_unit, copy=False)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat("pulse").hist().data,
        sc.array(
            dims=["detector"],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype="float64",
        ).fold(dim="detector", sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    workflow[time_of_flight.RawData] = raw_data
    workflow[time_of_flight.SimulationResults] = simulation_dream_choppers_time_overlap
    workflow[time_of_flight.LookupTableRelativeErrorThreshold] = 0.01
    workflow[time_of_flight.LtotalRange] = ltotal.min(), ltotal.max()

    # Make sure some values in the lookup table have been masked (turned to NaNs)
    original_table = workflow.compute(time_of_flight.TimeOfFlightLookupTable)
    masked_table = workflow.compute(time_of_flight.MaskedTimeOfFlightLookupTable)
    assert sc.isnan(masked_table).data.sum() > sc.isnan(original_table).data.sum()

    # Compute time-of-flight
    tofs = workflow.compute(time_of_flight.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result = wav_wfm["pulse", i].flatten(to="detector")
        for j in range(len(result)):
            computed_wavelengths = result[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths[:-2],
                true_wavelengths["pulse", i][:-2],
                rtol=sc.scalar(1e-02),
            )
            # The two neutrons in the overlap region should have NaN wavelengths
            assert sc.isnan(computed_wavelengths[-2])
            assert sc.isnan(computed_wavelengths[-1])


@pytest.fixture(scope="module")
def simulation_v20_choppers():
    return time_of_flight.simulate_beamline(
        choppers=fakes.wfm_choppers(), neutrons=300_000, seed=432
    )


@pytest.mark.parametrize("npulses", [1, 2])
@pytest.mark.parametrize(
    "ltotal",
    [
        sc.array(dims=["detector_number"], values=[26.0], unit="m"),
        sc.array(dims=["detector_number"], values=[26.0, 25.5], unit="m"),
        sc.array(
            dims=["y", "x"], values=[[26.0, 25.1, 26.33], [25.9, 26.0, 25.7]], unit="m"
        ),
    ],
)
@pytest.mark.parametrize("time_offset_unit", ["s", "ms", "us", "ns"])
@pytest.mark.parametrize("distance_unit", ["m", "mm"])
def test_v20_compute_wavelengths_from_wfm(
    simulation_v20_choppers, npulses, ltotal, time_offset_unit, distance_unit
):
    monitors = {
        f"detector{i}": ltot for i, ltot in enumerate(ltotal.flatten(to="detector"))
    }

    # Create some neutron events
    wavelengths = sc.array(
        dims=["event"], values=[2.75, 4.2, 5.4, 6.5, 7.6, 8.75], unit="angstrom"
    )
    birth_times = sc.full(sizes=wavelengths.sizes, value=1.5, unit="ms")
    ess_beamline = fakes.FakeBeamline(
        choppers=fakes.wfm_choppers(),
        monitors=monitors,
        run_length=sc.scalar(1 / 14, unit="s") * npulses,
        events_per_pulse=len(wavelengths),
        source=partial(
            tof_pkg.Source.from_neutrons,
            birth_times=birth_times,
            wavelengths=wavelengths,
            frequency=sc.scalar(14.0, unit="Hz"),
        ),
    )

    # Save the true wavelengths for later
    true_wavelengths = ess_beamline.source.data.coords["wavelength"]

    raw_data = sc.concat(
        [ess_beamline.get_monitor(key)[0] for key in monitors.keys()],
        dim="detector",
    ).fold(dim="detector", sizes=ltotal.sizes)

    # Convert the time offset to the unit requested by the test
    raw_data.bins.coords["event_time_offset"] = raw_data.bins.coords[
        "event_time_offset"
    ].to(unit=time_offset_unit, copy=False)

    raw_data.coords["Ltotal"] = ltotal.to(unit=distance_unit, copy=False)

    # Verify that all 6 neutrons made it through the chopper cascade
    assert sc.identical(
        raw_data.bins.concat("pulse").hist().data,
        sc.array(
            dims=["detector"],
            values=[len(wavelengths) * npulses] * len(monitors),
            unit="counts",
            dtype="float64",
        ).fold(dim="detector", sizes=ltotal.sizes),
    )

    # Set up the workflow
    workflow = sl.Pipeline(
        time_of_flight.providers(), params=time_of_flight.default_parameters()
    )

    workflow[time_of_flight.RawData] = raw_data
    workflow[time_of_flight.SimulationResults] = simulation_v20_choppers
    workflow[time_of_flight.LtotalRange] = ltotal.min(), ltotal.max()

    # Compute time-of-flight
    tofs = workflow.compute(time_of_flight.TofData)
    assert {dim: tofs.sizes[dim] for dim in ltotal.sizes} == ltotal.sizes

    # Convert to wavelength
    graph = {**beamline(scatter=False), **elastic("tof")}
    wav_wfm = tofs.transform_coords("wavelength", graph=graph)

    # Compare the computed wavelengths to the true wavelengths
    for i in range(npulses):
        result = wav_wfm["pulse", i].flatten(to="detector")
        for j in range(len(result)):
            computed_wavelengths = result[j].values.coords["wavelength"]
            assert sc.allclose(
                computed_wavelengths,
                true_wavelengths["pulse", i],
                rtol=sc.scalar(1e-02),
            )



import argparse
import shutil

import h5py


def _scale_group(event_data: h5py.Group, scale: int):
    if not all(
        required_field in event_data
        for required_field in ('event_index', 'event_time_offset', 'event_id')
    ):
        return
    event_index = (event_data['event_index'][:] * scale).astype('uint')
    event_data['event_index'][:] = event_index

    size = event_data['event_id'].size
    event_data['event_id'].resize(size * scale, axis=0)
    event_data['event_time_offset'].resize(size * scale, axis=0)

    for s in range(1, scale):
        event_data['event_id'][s * size : (s + 1) * size] = event_data['event_id'][
            :size
        ]
        event_data['event_time_offset'][s * size : (s + 1) * size] = event_data[
            'event_time_offset'
        ][:size]


def _grow_nexus_file_impl(file: h5py.File, detector_scale: int, monitor_scale: int):
    for group in file.values():
        if group.attrs.get('NX_class', '') == 'NXentry':
            entry = group
            break
    for group in entry.values():
        if group.attrs.get('NX_class', '') == 'NXinstrument':
            instrument = group
            break
    for group in instrument.values():
        if (nx_class := group.attrs.get('NX_class', '')) in (
            'NXdetector',
            'NXmonitor',
        ):
            for subgroup in group.values():
                if subgroup.attrs.get('NX_class', '') == 'NXevent_data':
                    _scale_group(
                        subgroup,
                        scale=detector_scale
                        if nx_class == 'NXdetector'
                        else monitor_scale,
                    )


def grow_nexus_file(*, filename: str, detector_scale: int, monitor_scale: int | None):
    with h5py.File(filename, 'a') as f:
        _grow_nexus_file_impl(
            f,
            detector_scale,
            monitor_scale if monitor_scale is not None else detector_scale,
        )


def integer_greater_than_one(x):
    x = int(x)
    if x < 1:
        raise argparse.ArgumentTypeError('Must be larger than or equal to 1')
    return x


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--file",
        type=str,
        help=(
            'Input file name. The events in the input file will be '
            'repeated `scale` times and stored in the output file.'
        ),
        required=True,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help='Output file name where the resulting nexus file will be written.',
        required=True,
    )
    parser.add_argument(
        "-s",
        "--detector-scale",
        type=integer_greater_than_one,
        help=('Scale factor to multiply the number of detector events by.'),
        required=True,
    )
    parser.add_argument(
        "-m",
        "--monitor-scale",
        type=integer_greater_than_one,
        default=None,
        help=(
            'Scale factor to multiply the number of monitor events by. '
            'If not given, the detector scale will be used'
        ),
    )
    args = parser.parse_args()
    if args.file != args.output:
        shutil.copy2(args.file, args.output)
    grow_nexus_file(
        filename=args.output,
        detector_scale=args.detector_scale,
        monitor_scale=args.monitor_scale,
    )



_spinner = '''
<style>
.lds-spinner {
  /* change color here */
  color: #1c4c5b
}
.lds-spinner,
.lds-spinner div,
.lds-spinner div:after {
  box-sizing: border-box;
}
.lds-spinner {
  color: currentColor;
  display: inline-block;
  position: relative;
  width: 80px;
  height: 80px;
}
.lds-spinner div {
  transform-origin: 40px 40px;
  animation: lds-spinner 1.2s linear infinite;
}
.lds-spinner div:after {
  content: " ";
  display: block;
  position: absolute;
  top: 3.2px;
  left: 36.8px;
  width: 6.4px;
  height: 17.6px;
  border-radius: 20%;
  background: currentColor;
}
.lds-spinner div:nth-child(1) {
  transform: rotate(0deg);
  animation-delay: -1.1s;
}
.lds-spinner div:nth-child(2) {
  transform: rotate(30deg);
  animation-delay: -1s;
}
.lds-spinner div:nth-child(3) {
  transform: rotate(60deg);
  animation-delay: -0.9s;
}
.lds-spinner div:nth-child(4) {
  transform: rotate(90deg);
  animation-delay: -0.8s;
}
.lds-spinner div:nth-child(5) {
  transform: rotate(120deg);
  animation-delay: -0.7s;
}
.lds-spinner div:nth-child(6) {
  transform: rotate(150deg);
  animation-delay: -0.6s;
}
.lds-spinner div:nth-child(7) {
  transform: rotate(180deg);
  animation-delay: -0.5s;
}
.lds-spinner div:nth-child(8) {
  transform: rotate(210deg);
  animation-delay: -0.4s;
}
.lds-spinner div:nth-child(9) {
  transform: rotate(240deg);
  animation-delay: -0.3s;
}
.lds-spinner div:nth-child(10) {
  transform: rotate(270deg);
  animation-delay: -0.2s;
}
.lds-spinner div:nth-child(11) {
  transform: rotate(300deg);
  animation-delay: -0.1s;
}
.lds-spinner div:nth-child(12) {
  transform: rotate(330deg);
  animation-delay: 0s;
}
@keyframes lds-spinner {
  0% {
    opacity: 1;
  }
  100% {
    opacity: 0;
  }
}
</style>
<div class="lds-spinner"><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div><div></div></div>
'''  # noqa: E501


class Spinner:
    def _repr_html_(self, *args, **kwargs):
        return _spinner

    def _repr_pretty_(self, *args, **kwargs):
        return '...in progress'



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
from collections.abc import Mapping

import scipp as sc
from scippneutron.chopper import DiskChopper

from .types import SimulationResults


def simulate_beamline(
    choppers: Mapping[str, DiskChopper],
    neutrons: int = 1_000_000,
    seed: int | None = None,
    facility: str = 'ess',
) -> SimulationResults:
    """
    Simulate a pulse of neutrons propagating through a chopper cascade using the
    ``tof`` package (https://tof.readthedocs.io).

    Parameters
    ----------
    choppers:
        A dict of DiskChopper objects representing the choppers in the beamline. See
        https://scipp.github.io/scippneutron/user-guide/chopper/processing-nexus-choppers.html#Build-DiskChopper
        for more information.
    neutrons:
        Number of neutrons to simulate.
    seed:
        Seed for the random number generator used in the simulation.
    facility:
        Facility where the experiment is performed.
    """
    import tof

    tof_choppers = [
        tof.Chopper(
            frequency=abs(ch.frequency),
            direction=tof.AntiClockwise
            if (ch.frequency.value > 0.0)
            else tof.Clockwise,
            open=ch.slit_begin,
            close=ch.slit_end,
            phase=abs(ch.phase),
            distance=ch.axle_position.fields.z,
            name=name,
        )
        for name, ch in choppers.items()
    ]
    source = tof.Source(facility=facility, neutrons=neutrons, seed=seed)
    if not tof_choppers:
        events = source.data.squeeze()
        return SimulationResults(
            time_of_arrival=events.coords["time"],
            speed=events.coords["speed"],
            wavelength=events.coords["wavelength"],
            weight=events.data,
            distance=0.0 * sc.units.m,
        )
    model = tof.Model(source=source, choppers=tof_choppers)
    results = model.run()
    # Find name of the furthest chopper in tof_choppers
    furthest_chopper = max(tof_choppers, key=lambda c: c.distance)
    events = results[furthest_chopper.name].data.squeeze()
    events = events[
        ~(events.masks["blocked_by_others"] | events.masks["blocked_by_me"])
    ]
    return SimulationResults(
        time_of_arrival=events.coords["toa"],
        speed=events.coords["speed"],
        wavelength=events.coords["wavelength"],
        weight=events.data,
        distance=furthest_chopper.distance,
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scippnexus as snx
from scipp.testing import assert_identical

from ess.reduce import data
from ess.reduce.nexus.json_generator import event_data_generator
from ess.reduce.nexus.json_nexus import json_nexus_group


def test_event_data_generator_monitor_events_round_trip() -> None:
    filename = data.loki_tutorial_sample_run_60250()
    monitor = snx.load(filename, root='entry/instrument/monitor_1/monitor_1_events')
    generator = event_data_generator(monitor)
    for i in range(len(monitor)):
        group = json_nexus_group(next(generator))
        assert_identical(group[()], monitor[i : i + 1])
    with pytest.raises(StopIteration):
        next(generator)


def test_event_data_generator_detector_events_round_trip() -> None:
    filename = data.loki_tutorial_sample_run_60250()
    detector = snx.load(
        filename, root='entry/instrument/larmor_detector/larmor_detector_events'
    )
    generator = event_data_generator(detector)
    for i in range(100):
        group = json_nexus_group(next(generator))
        assert_identical(group[()], detector[i : i + 1])


def test_event_data_generator_without_event_id_yields_ones() -> None:
    filename = data.loki_tutorial_sample_run_60250()
    base = snx.load(filename, root='entry/instrument/monitor_1/monitor_1_events')
    monitor = base.bins.drop_coords('event_id')
    generator = event_data_generator(monitor)
    for i in range(100):
        group = json_nexus_group(next(generator))
        assert_identical(group[()], base[i : i + 1])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from __future__ import annotations

from collections.abc import Callable, MutableSet, Sequence
from typing import Any, TypeVar

import networkx as nx
from sciline import Pipeline
from sciline._utils import key_name
from sciline.typing import Key

from .parameter import Parameter, keep_default, parameter_mappers, parameter_registry

T = TypeVar("T")


class WorkflowRegistry(MutableSet):
    def __init__(self):
        self._workflows: dict[str, type] = {}

    def __contains__(self, item: object) -> bool:
        return item in self._workflows.values()

    def __iter__(self):
        return iter(self._workflows.values())

    def __len__(self) -> int:
        return len(self._workflows)

    def add(self, value: type) -> None:
        key = value.__qualname__
        self._workflows[key] = value

    def discard(self, value: type) -> None:
        self._workflows = {k: v for k, v in self._workflows.items() if v != value}


workflow_registry = WorkflowRegistry()


def register_workflow(cls: Callable[[], Pipeline]) -> Callable[[], Pipeline]:
    workflow_registry.add(cls)
    return cls


def _get_defaults_from_workflow(workflow: Pipeline) -> dict[Key, Any]:
    nodes = workflow.underlying_graph.nodes
    return {key: values["value"] for key, values in nodes.items() if "value" in values}


def get_typical_outputs(pipeline: Pipeline) -> tuple[Key, ...]:
    if (typical_outputs := getattr(pipeline, "typical_outputs", None)) is None:
        graph = pipeline.underlying_graph
        sink_nodes = [node for node, degree in graph.out_degree if degree == 0]
        return sorted(_with_pretty_names(sink_nodes))
    return _with_pretty_names(typical_outputs)


def get_possible_outputs(pipeline: Pipeline) -> tuple[Key, ...]:
    return sorted(_with_pretty_names(tuple(pipeline.underlying_graph.nodes)))


def _with_pretty_names(outputs: Sequence[Key]) -> tuple[tuple[str, Key], ...]:
    """Add a more readable string representation without full module path."""
    return tuple((key_name(output), output) for output in outputs)


def get_parameters(
    pipeline: Pipeline, outputs: tuple[Key, ...]
) -> dict[Key, Parameter]:
    """Return a dictionary of parameters for the workflow."""
    subgraph = set(outputs)
    graph = pipeline.underlying_graph
    for key in outputs:
        subgraph.update(nx.ancestors(graph, key))
    defaults = _get_defaults_from_workflow(pipeline)
    return {
        key: param.with_default(defaults.get(key, keep_default))
        for key, param in parameter_registry.items()
        if key in subgraph
    }


def assign_parameter_values(pipeline: Pipeline, values: dict[Key, Any]) -> Pipeline:
    """Set a value for a parameter in the pipeline."""
    pipeline = pipeline.copy()
    for key, value in values.items():
        if (mapper := parameter_mappers.get(key)) is not None:
            pipeline = mapper(pipeline, value)
        else:
            pipeline[key] = value
    return pipeline



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)

from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from typing import Any

import numpy as np
import pytest
import scipp as sc
import scipp.testing
import scippnexus as snx

from ess.reduce import nexus
from ess.reduce.nexus.types import NeXusLocationSpec

year_zero = sc.datetime('1970-01-01T00:00:00')


def _event_data_components() -> sc.DataGroup:
    return sc.DataGroup(
        {
            'event_id': sc.array(dims=['event'], unit=None, values=[1, 2, 4, 1, 2, 2]),
            'event_time_offset': sc.array(
                dims=['event'], unit='s', values=[456, 7, 3, 345, 632, 23]
            ),
            'event_time_zero': sc.epoch(unit='s')
            + sc.array(dims=['event_index'], unit='s', values=[1, 2, 3, 4]),
            'event_index': sc.array(
                dims=['event_index'], unit=None, values=[0, 3, 3, 6]
            ),
            'detector_number': sc.arange('detector_number', 5, unit=None),
            'pixel_offset': sc.vectors(
                dims=['detector_number'],
                values=np.arange(3 * 5).reshape((5, 3)),
                unit='m',
            ),
        }
    )


def detector_transformation_components() -> sc.DataGroup:
    return sc.DataGroup(
        {
            'offset': sc.vector([0.4, 0.0, 11.5], unit='m'),
        }
    )


def _monitor_histogram() -> sc.DataArray:
    return sc.DataArray(
        sc.array(dims=['time'], values=[2, 4, 8, 3], unit='counts'),
        coords={
            'time': sc.epoch(unit='ms')
            + sc.array(dims=['time'], values=[2, 4, 6, 8, 10], unit='ms'),
        },
    )


def _source_data() -> sc.DataGroup:
    return sc.DataGroup(
        {
            'name': 'moderator',
            'probe': 'neutron',
            'type': 'Spallation Neutron Source',
            'position': sc.vector([0, 0, 0], unit='m'),
            'transform': sc.spatial.translation(value=[0, 0, 0], unit='m'),
            'nexus_component_name': 'source',
        }
    )


def _sample_data() -> sc.DataGroup:
    return sc.DataGroup(
        {
            'name': 'water',
            'chemical_formula': 'H2O',
            'type': 'sample+can',
            'nexus_component_name': 'sample',
        }
    )


def _choppers_data() -> sc.DataGroup[sc.DataGroup[Any]]:
    return sc.DataGroup[sc.DataGroup](
        {
            'chopper_1': sc.DataGroup[Any](
                {
                    'slit_edges': sc.array(dims=['dim_0'], values=[-5, 45], unit='deg'),
                    'slit_height': sc.scalar(0.054, unit='m'),
                    'slits': np.int64(1),  # snx returns this type when loading
                    'nexus_component_name': 'chopper_1',
                }
            ),
            'chopper_2': sc.DataGroup[Any](
                {
                    'slit_edges': sc.array(dims=['dim_0'], values=[15, 60], unit='deg'),
                    'slit_height': sc.scalar(0.07, unit='m'),
                    'slits': np.int64(1),
                    'nexus_component_name': 'chopper_2',
                }
            ),
        }
    )


def _analyzers_data() -> sc.DataGroup[sc.DataGroup[Any]]:
    return sc.DataGroup[sc.DataGroup[Any]](
        {
            'analyzer_B': sc.DataGroup[Any](
                {
                    'd_spacing': sc.scalar(3.355, unit='angstrom'),
                    'usage': 'Bragg',
                    'nexus_component_name': 'analyzer_B',
                }
            ),
            'analyzer_A': sc.DataGroup[Any](
                {
                    'd_spacing': sc.scalar(3.104, unit='angstrom'),
                    'usage': 'Bragg',
                    'nexus_component_name': 'analyzer_A',
                }
            ),
        }
    )


def _write_transformation(group: snx.Group, offset: sc.Variable) -> None:
    group.create_field('depends_on', sc.scalar('transformations/t1'))
    transformations = group.create_class('transformations', snx.NXtransformations)
    t1 = transformations.create_field('t1', sc.scalar(0.0, unit=offset.unit))
    t1.attrs['depends_on'] = '.'
    t1.attrs['transformation_type'] = 'translation'
    t1.attrs['offset'] = offset.values
    t1.attrs['offset_units'] = str(offset.unit)
    t1.attrs['vector'] = sc.vector([0, 0, 1]).value


def _write_nexus_data(store: Path | BytesIO) -> None:
    with snx.File(store, 'w') as root:
        entry = root.create_class('entry-001', snx.NXentry)
        instrument = entry.create_class('reducer', snx.NXinstrument)

        detector = instrument.create_class('bank12', snx.NXdetector)
        events = detector.create_class('bank12_events', snx.NXevent_data)
        detector_components = _event_data_components()
        events['event_id'] = detector_components['event_id']
        events['event_time_offset'] = detector_components['event_time_offset']
        events['event_time_zero'] = detector_components['event_time_zero']
        events['event_index'] = detector_components['event_index']
        detector['x_pixel_offset'] = detector_components['pixel_offset'].fields.x
        detector['y_pixel_offset'] = detector_components['pixel_offset'].fields.y
        detector['z_pixel_offset'] = detector_components['pixel_offset'].fields.z
        detector['detector_number'] = detector_components['detector_number']
        _write_transformation(detector, detector_transformation_components()['offset'])

        monitor_data = _monitor_histogram()
        monitor = instrument.create_class('monitor', snx.NXmonitor)
        data = monitor.create_class('data', snx.NXdata)
        signal = data.create_field('signal', monitor_data.data)
        signal.attrs['signal'] = 1
        signal.attrs['axes'] = monitor_data.dim
        data.create_field('time', monitor_data.coords['time'])

        source_data = _source_data()
        source = instrument.create_class('source', snx.NXsource)
        source.create_field('name', source_data['name'])
        source.create_field('probe', source_data['probe'])
        source.create_field('type', source_data['type'])
        _write_transformation(source, source_data['position'])

        sample_data = _sample_data()
        sample = entry.create_class('sample', snx.NXsample)
        sample.create_field('name', sample_data['name'])
        sample.create_field('chemical_formula', sample_data['chemical_formula'])
        sample.create_field('type', sample_data['type'])

        for name, chopper in _choppers_data().items():
            chop = instrument.create_class(name, snx.NXdisk_chopper)
            chop.create_field('slit_edges', chopper['slit_edges'])
            chop.create_field('slit_height', chopper['slit_height'])
            chop.create_field('slits', chopper['slits'])

        for name, analyzer in _analyzers_data().items():
            ana = instrument.create_class(name, snx.NXcrystal)
            ana.create_field('d_spacing', analyzer['d_spacing'])
            ana.create_field('usage', analyzer['usage'])


@contextmanager
def _file_store(request: pytest.FixtureRequest):
    if request.param == BytesIO:
        yield BytesIO()
    else:
        # It would be good to use pyfakefs here, but h5py
        # uses C to open files and that bypasses the fake.
        base = request.getfixturevalue('tmp_path')
        yield base / 'testfile.nxs'


@pytest.fixture(params=[Path, BytesIO, snx.Group])
def nexus_file(request):
    with _file_store(request) as store:
        _write_nexus_data(store)
        if isinstance(store, BytesIO):
            store.seek(0)

        if request.param in (Path, BytesIO):
            yield store
        else:
            with snx.File(store, 'r') as f:
                yield f


@pytest.fixture
def expected_bank12():
    components = _event_data_components()
    buffer = sc.DataArray(
        sc.ones(sizes={'event': 6}, unit='counts', dtype='float32'),
        coords={
            'detector_number': components['event_id'],
            'event_time_offset': components['event_time_offset'],
        },
    )

    # Bin by event_index tp broadcast event_time_zero to events
    binned_in_time = sc.DataArray(
        sc.bins(
            data=buffer,
            begin=components['event_index'],
            end=sc.concat(
                [components['event_index'][1:], components['event_index'][-1]],
                dim='event_index',
            ),
            dim='event',
        )
    )
    binned_in_time.bins.coords['event_time_zero'] = sc.bins_like(
        binned_in_time, components['event_time_zero']
    )

    # Bin by detector number like ScippNexus would
    binned = binned_in_time.bins.concat().group(components['detector_number'])
    binned.coords['x_pixel_offset'] = components['pixel_offset'].fields.x
    binned.coords['y_pixel_offset'] = components['pixel_offset'].fields.y
    binned.coords['z_pixel_offset'] = components['pixel_offset'].fields.z
    # Computed position
    offset = detector_transformation_components()['offset']
    binned.coords['position'] = offset + components['pixel_offset']
    return binned


@pytest.fixture
def expected_monitor() -> sc.DataArray:
    return _monitor_histogram()


@pytest.fixture
def expected_source() -> sc.DataGroup:
    return _source_data()


@pytest.fixture
def expected_sample() -> sc.DataGroup:
    return _sample_data()


@pytest.fixture
def expected_choppers() -> sc.DataGroup[sc.DataGroup[Any]]:
    return _choppers_data()


@pytest.fixture
def expected_analyzers() -> sc.DataGroup[sc.DataGroup[Any]]:
    return _analyzers_data()


def test_load_data_loads_expected_event_data(nexus_file, expected_bank12):
    events = nexus.load_data(
        nexus_file,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    grouped = nexus.group_event_data(
        event_data=events,
        detector_number=expected_bank12.coords['detector_number'],
    )
    expected = expected_bank12.drop_coords(
        ['position', 'x_pixel_offset', 'y_pixel_offset', 'z_pixel_offset']
    )
    sc.testing.assert_identical(grouped, expected)


def test_load_data_loads_expected_histogram_data(nexus_file, expected_monitor):
    histogram = nexus.load_data(
        nexus_file,
        component_name=nexus.types.NeXusName[nexus.types.Monitor1]('monitor'),
    )
    sc.testing.assert_identical(histogram, expected_monitor)


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
@pytest.mark.parametrize(
    'selection',
    [
        None,
        ('event_time_zero', slice(year_zero, None)),
        ('event_time_zero', slice(None, year_zero)),
    ],
)
def test_load_detector(nexus_file, expected_bank12, entry_name, selection):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    if selection is not None:
        loc.selection = selection
    detector = nexus.load_component(loc, nx_class=snx.NXdetector)
    detector = nexus.compute_component_position(detector)
    if selection:
        expected = expected_bank12.bins[selection]
        expected.coords.pop(selection[0])
    else:
        expected = expected_bank12
    sc.testing.assert_identical(detector['bank12_events'], expected)
    offset = detector_transformation_components()['offset']
    sc.testing.assert_identical(
        detector['transform'],
        sc.spatial.translation(unit=offset.unit, value=offset.value),
    )


@pytest.mark.parametrize(
    'selection',
    [
        (),
        {'event_time_zero': slice(2, None)},
        {'event_time_zero': slice(None, 3)},
        {'event_time_zero': slice(1, 3)},
    ],
)
def test_load_and_group_event_data_consistent_with_load_via_detector(
    nexus_file, selection
):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    if selection:
        loc.selection = selection
    detector = nexus.load_component(loc, nx_class=snx.NXdetector)
    detector = nexus.compute_component_position(detector)['bank12_events']
    events = nexus.load_data(
        nexus_file,
        selection=selection,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    grouped = nexus.group_event_data(
        event_data=events,
        detector_number=detector.coords['detector_number'],
    )
    scipp.testing.assert_identical(detector.data, grouped.data)


def test_group_event_data_does_not_modify_input(nexus_file):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    detector = nexus.load_component(loc, nx_class=snx.NXdetector)
    detector = nexus.compute_component_position(detector)['bank12_events']
    events = nexus.load_data(
        nexus_file,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    _ = nexus.group_event_data(
        event_data=events,
        detector_number=detector.coords['detector_number'],
    )
    assert 'event_time_zero' not in events.bins.coords


def test_load_detector_open_file_with_new_definitions_raises(nexus_file):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        component_name=nexus.types.NeXusDetectorName('bank12'),
    )
    if isinstance(nexus_file, snx.Group):
        with pytest.raises(ValueError, match="new definitions"):
            nexus.load_component(loc, nx_class=snx.NXdetector, definitions={})
        # Passing same definitions should work
        nexus.load_component(
            loc, nx_class=snx.NXdetector, definitions=nexus_file._definitions.copy()
        )
    else:
        nexus.load_component(loc, nx_class=snx.NXdetector, definitions={})


def test_load_detector_new_definitions_applied(nexus_file):
    if not isinstance(nexus_file, snx.Group):
        new_definition_used = False

        def detector(*args, **kwargs):
            nonlocal new_definition_used
            new_definition_used = True
            return snx.base_definitions()['NXdetector'](*args, **kwargs)

        loc = NeXusLocationSpec(
            filename=nexus_file,
            component_name=nexus.types.NeXusDetectorName('bank12'),
        )

        nexus.load_component(
            loc,
            nx_class=snx.NXdetector,
            definitions=dict(snx.base_definitions(), NXdetector=detector),
        )
        assert new_definition_used


def test_load_detector_requires_entry_name_if_not_unique(nexus_file):
    if not isinstance(nexus_file, Path):
        # For simplicity, only create a second entry in an actual file
        return

    with snx.File(nexus_file, 'r+') as f:
        f.create_class('entry', snx.NXentry)

    loc = NeXusLocationSpec(
        filename=nexus.types.FilePath(nexus_file),
        component_name=nexus.types.NeXusDetectorName('bank12'),
        entry_name=None,
    )
    with pytest.raises(ValueError, match="Expected exactly one"):
        nexus.load_component(loc, nx_class=snx.NXdetector)


def test_load_detector_select_entry_if_not_unique(nexus_file, expected_bank12):
    if not isinstance(nexus_file, Path):
        # For simplicity, only create a second entry in an actual file
        return

    with snx.File(nexus_file, 'r+') as f:
        f.create_class('entry', snx.NXentry)

    loc = NeXusLocationSpec(
        filename=nexus.types.FilePath(nexus_file),
        component_name=nexus.types.NeXusDetectorName('bank12'),
        entry_name=nexus.types.NeXusEntryName('entry-001'),
    )
    detector = nexus.load_component(loc, nx_class=snx.NXdetector)
    detector = nexus.compute_component_position(detector)
    sc.testing.assert_identical(detector['bank12_events'], expected_bank12)


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
@pytest.mark.parametrize(
    'selection',
    [
        None,
        ('time', slice(year_zero.to(unit='ms'), None)),
        ('time', slice(None, year_zero.to(unit='ms'))),
    ],
)
def test_load_monitor(nexus_file, expected_monitor, entry_name, selection):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
        component_name=nexus.types.NeXusName[nexus.types.Monitor1]('monitor'),
    )
    if selection is not None:
        loc.selection = selection
    monitor = nexus.load_component(loc, nx_class=snx.NXmonitor)
    monitor = nexus.compute_component_position(monitor)
    expected = expected_monitor[selection] if selection else expected_monitor
    sc.testing.assert_identical(monitor['data'], expected)


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
@pytest.mark.parametrize('source_name', [None, nexus.types.NeXusSourceName('source')])
def test_load_source(nexus_file, expected_source, entry_name, source_name):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
        component_name=source_name,
    )
    source = nexus.load_component(loc, nx_class=snx.NXsource)
    source = nexus.compute_component_position(source)
    # NeXus details that we don't need to test as long as the positions are ok:
    del source['depends_on']
    del source['transformations']
    sc.testing.assert_identical(source, expected_source)


@pytest.mark.parametrize(
    ('loader', 'cls', 'name'),
    [
        (nexus.load_component, snx.NXsource, 'NXsource'),
        (nexus.load_component, snx.NXsample, 'NXsample'),
    ],
)
def test_load_new_definitions_applied(nexus_file, loader, cls, name):
    if not isinstance(nexus_file, snx.Group):
        new_definition_used = False

        def new(*args, **kwargs):
            nonlocal new_definition_used
            new_definition_used = True
            return cls(*args, **kwargs)

        loc = NeXusLocationSpec(filename=nexus_file)
        loader(loc, nx_class=cls, definitions={**snx.base_definitions(), name: new})
        assert new_definition_used


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
def test_load_sample(nexus_file, expected_sample, entry_name):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
    )
    sample = nexus.load_component(loc, nx_class=snx.NXsample)
    sc.testing.assert_identical(sample, expected_sample)


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
def test_load_disk_choppers(nexus_file, expected_choppers, entry_name):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
    )
    choppers = nexus.load_all_components(loc, nx_class=snx.NXdisk_chopper)
    sc.testing.assert_identical(choppers, expected_choppers)


@pytest.mark.parametrize('entry_name', [None, nexus.types.NeXusEntryName('entry-001')])
def test_load_analyzers(nexus_file, expected_analyzers, entry_name):
    loc = NeXusLocationSpec(
        filename=nexus_file,
        entry_name=entry_name,
    )
    analyzers = nexus.load_all_components(loc, nx_class=snx.NXcrystal)
    sc.testing.assert_identical(analyzers, expected_analyzers)


def test_extract_detector_data():
    detector = sc.DataGroup(
        {
            'jdl2ab': sc.data.binned_x(10, 3),
            'llk': 23,
            ' _': sc.linspace('xx', 2, 3, 10),
        }
    )
    data = nexus.extract_signal_data_array(detector)
    sc.testing.assert_identical(data, detector['jdl2ab'])


def test_extract_monitor_data():
    monitor = sc.DataGroup(
        {
            '(eed)': sc.data.data_xy(),
            'llk': 23,
            ' _': sc.linspace('xx', 2, 3, 10),
        }
    )
    data = nexus.extract_signal_data_array(monitor)
    sc.testing.assert_identical(data, monitor['(eed)'])


def test_extract_detector_data_requires_unique_dense_data():
    detector = sc.DataGroup(
        {
            'jdl2ab': sc.data.data_xy(),
            'llk': 23,
            'lob': sc.data.data_xy(),
            ' _': sc.linspace('xx', 2, 3, 10),
        }
    )
    with pytest.raises(
        ValueError, match="Cannot uniquely identify the data to extract"
    ):
        nexus.extract_signal_data_array(detector)


def test_extract_detector_data_ignores_position_data_array():
    detector = sc.DataGroup(jdl2ab=sc.data.data_xy(), position=sc.data.data_xy())
    nexus.extract_signal_data_array(detector)


def test_extract_detector_data_ignores_transform_data_array():
    detector = sc.DataGroup(jdl2ab=sc.data.data_xy(), transform=sc.data.data_xy())
    nexus.extract_signal_data_array(detector)


def test_extract_detector_data_requires_unique_event_data():
    detector = sc.DataGroup(
        {
            'jdl2ab': sc.data.binned_x(10, 3),
            'llk': 23,
            'lob': sc.data.binned_x(14, 5),
            ' _': sc.linspace('xx', 2, 3, 10),
        }
    )
    with pytest.raises(
        ValueError, match="Cannot uniquely identify the data to extract"
    ):
        nexus.extract_signal_data_array(detector)


def test_extract_detector_data_favors_event_data_over_histogram_data():
    detector = sc.DataGroup(
        {
            'jdl2ab': sc.data.data_xy(),
            'llk': 23,
            'lob': sc.data.binned_x(14, 5),
            ' _': sc.linspace('xx', 2, 3, 10),
        }
    )
    data = nexus.extract_signal_data_array(detector)
    sc.testing.assert_identical(data, detector['lob'])


def compute_component_position_returns_input_if_no_depends_on() -> None:
    dg = sc.DataGroup(position=sc.vector([1, 2, 3], unit='m'))
    result = nexus.compute_component_position(dg)
    assert result is dg
